# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

import numpy as np
try:
    from shapely.geometry import LineString, Point
    from shapely.geometry.polygon import Polygon
    shapely_imported = True
except ImportError:
    shapely_imported = False

from pandapower.auxiliary import soft_dependency_error
from pandapower.control.controller.DERController.DERBasics import BaseModel

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)

# -------------------------------------------------------------------------------------------------
""" This file defines PQ, QV, and PQV areas of reactive power provision capabilities including those
defined in VDE-AR-N technical connection rule standards"""
# -------------------------------------------------------------------------------------------------

""" Base class """

class BaseArea(BaseModel):
    """ Defines which functionality is common to all area (PQ or QV) classes. """
    def in_area(self, p_pu, q_pu, vm_pu):
        min_max_q = self.q_flexibility(p_pu, vm_pu)
        return (min_max_q[:, 0] <= q_pu) & (min_max_q[:, 1] >= q_pu)

    def q_flexibility(self, p_pu, vm_pu):
        pass

    def __str__(self):
        return self.__class__.name()


class BasePQVArea(BaseArea):
    """ Defines functionality common for mulitple PQVArea classes. """
    def __init__(self, raise_merge_overlap=True):
        self.raise_merge_overlap = raise_merge_overlap

    def in_area(self, p_pu, q_pu, vm_pu):
        return self.pq_area.in_area(p_pu, q_pu, vm_pu) & self.qv_area.in_area(p_pu, q_pu, vm_pu)

    def q_flexibility(self, p_pu, vm_pu):
        # PQ area flexibility
        min_max_q = self.pq_area.q_flexibility(p_pu)
        no_flexibility = min_max_q[:, 0] > min_max_q[:, 1]
        if n_no_flex := sum(no_flexibility):
            raise ValueError(f"For {n_no_flex} elements, the q flexibility of the PQ area provides "
                             "max_q < min_q.")

        # QV area flexibility
        min_max_q_qv = self.qv_area.q_flexibility(None, vm_pu)
        no_flexibility = min_max_q_qv[:, 0] > min_max_q_qv[:, 1]
        if n_no_flex := sum(no_flexibility):
            raise ValueError(f"For {n_no_flex} elements, the q flexibility of the QV area provides "
                             "max_q < min_q.")

        # merge q flexibility of pq and qv areas
        min_max_q[:, 0] = np.maximum(min_max_q[:, 0], min_max_q_qv[:, 0])
        min_max_q[:, 1] = np.minimum(min_max_q[:, 1], min_max_q_qv[:, 1])

        no_flexibility = min_max_q[:, 0] > min_max_q[:, 1]
        if n_no_flex := sum(no_flexibility):
            error_msg = (f"For {n_no_flex} elements, max_q > min_q result from considering both "
                         "the PQ area flexibility and the QV area flexibility.")
            if self.raise_merge_overlap:
                raise ValueError(error_msg)
            else:
                logger.error(error_msg)
            min_max_q[no_flexibility, 0] = np.sum(min_max_q[no_flexibility], axis=1)/2
            min_max_q[no_flexibility, 1] = min_max_q[no_flexibility, 0]

        return min_max_q


""" Polygon Areas """

class PQAreaPOLYGON(BaseArea):
    """ Provides a polygonal area of feasible reactive power provision. The polygonal area can be
    defined by 'p_points_pu' and 'q_points_pu'.

    Note: Due to generator point of view, negative q values are correspond with underexcited behavior.

    Example
    -------
    >>> PQAreaDefault(p_points_pu=(0.1, 0.2, 1, 1, 0.2, 0.1, 0.1),
    ...               q_points_pu=(0.1, 0.410775, 0.410775, -0.328684, -0.328684, -0.1, 0.1))
    """
    def __init__(self, p_points_pu, q_points_pu):
        self.p_points_pu = p_points_pu
        self.q_points_pu = q_points_pu

        if not shapely_imported:
            soft_dependency_error("PQAreaDefault", "shapely")

        self.polygon = Polygon([(p_pu, q_pu) for p_pu, q_pu in zip(p_points_pu, q_points_pu)])

    def in_area(self, p_pu, q_pu, vm_pu=None):
        return np.array([self.polygon.contains(Point(pi, qi)) for pi, qi in zip(p_pu, q_pu)])

    def q_flexibility(self, p_pu, vm_pu=None):
        def _q_flex(p_pu):
            line = LineString([(p_pu, -1), (p_pu, 1)])
            if line.intersects(self.polygon):
                points = [point[1] for point in LineString(
                    [(p_pu, -1), (p_pu, 1)]).intersection(self.polygon).coords]
                if len(points) == 1:
                    return [points[0]]*2
                elif len(points) == 2:
                    return points
                else:
                    raise ValueError(f"{len(points)=} is wrong. 2 or 1 is expected.")
            else:
                return [0, 0]
        return np.r_[[_q_flex(pi) for pi in p_pu]]


class QVAreaPOLYGON(BaseArea):
    """ Provides a polygonal area of feasible reactive power provision. The polygonal area can be
    defined by 'q_points_pu' and 'vm_points_pu'.

    Note: Due to generator point of view, negative q values are correspond with underexcited behavior.

    Example
    -------
    >>> QVAreaPOLYGON(q_points_pu=(0.1, 0.410775, 0.410775, -0.328684, -0.328684, -0.1, 0.1),
    ...               vm_points_pu=(0.9, 1.05, 1.1, 1.1, 1.05, 0.9, 0.9))
    """
    def __init__(self, q_points_pu, vm_points_pu):
        self.q_points_pu = q_points_pu
        self.vm_points_pu = vm_points_pu

        if not shapely_imported:
            soft_dependency_error("PQAreaDefault", "shapely")

        # note that the naming QVAreaPOLYGON might be confusing because, in fact, it is a VQAreaPOLYGON
        self.polygon = Polygon([(vm, q) for vm, q in zip(vm_points_pu, q_points_pu)])

    def in_area(self, p_pu, q_pu, vm_pu):
        return np.array([self.polygon.contains(Point(vmi, qi)) for vmi, qi in zip(vm_pu, q_pu)])

    def q_flexibility(self, p_pu, vm_pu):
        assert all(vm_pu >= 0) and all(vm_pu <= 2)
        def _q_flex(vm_pu):
            line = LineString([(vm_pu, -1), (vm_pu, 1)])
            if line.intersects(self.polygon):
                return [point[1] for point in LineString(
                    [(vm_pu, -1), (vm_pu, 1)]).intersection(self.polygon).coords]
            else:
                return [0, 0]
        return np.r_[[_q_flex(vmi) for vmi in vm_pu]]


class PQVAreaPOLYGON(BasePQVArea):
    """ Provides a polygonal area of feasible reactive power provision. The polygonal area can be
    defined by 'p_points_pu' and 'q_pq_points_pu' as well as 'q_qv_points_pu' and 'vm_points_pu'.

    Example
    -------
    >>> PQVAreaDefault(p_points_pu=(0.1, 0.2, 1, 1, 0.2, 0.1, 0.1),
    ...                q_pq_points_pu=(0.1, 0.410775, 0.410775, -0.328684, -0.328684, -0.1, 0.1),
    ...                q_qv_points_pu=(0.1, 0.410775, 0.410775, -0.328684, -0.328684, -0.1, 0.1),
    ...                vm_points_pu=(0.9, 1.05, 1.1, 1.1, 1.05, 0.9, 0.9))
    """
    def __init__(self, p_points_pu, q_pq_points_pu, q_qv_points_pu, vm_points_pu,
                 raise_merge_overlap=True):
        super().__init__(raise_merge_overlap=raise_merge_overlap)
        self.pq_area = PQAreaPOLYGON(p_points_pu, q_pq_points_pu)
        self.qv_area = QVAreaPOLYGON(q_qv_points_pu, vm_points_pu)


""" STATCOM DERs: """


class PQAreaSTATCOM(BaseArea):
    """ PQAreaSTATCOM provides a simple rectangular reactive power provision area without a
    dependency on the on the active power.

    Example
    -------
    >>> PQAreaSTATCOM(min_q=-0.328684, max_q=0.410775)
    """

    def __init__(self, min_q_pu, max_q_pu):
        self.min_q_pu, self.max_q_pu = min_q_pu, max_q_pu

    def in_area(self, p_pu, q_pu, vm_pu=None):
        return (self.min_q_pu <= q_pu) & (q_pu <= self.max_q_pu)

    def q_flexibility(self, p_pu, vm_pu=None):
        return np.c_[[self.min_q_pu]*len(p_pu), [self.max_q_pu]*len(p_pu)]


""" HV DERs: """


class PQArea4120(BaseArea):
    """ This class models the PQ area of flexible Q for high-voltage plants according to
    VDE AR-N-4120. It is used to be combined with Voltage dependencies in PQVArea4120V1,
    PQVArea4120V2 and PQVArea4120V3.
    """
    def __init__(self, min_q_pu, max_q_pu, version=2018, **kwargs):
        supported_versions = [2015, 2018]
        p_points_pu = {2015: [0.1, 0.2], 2018: [0.05, 0.2]}
        if version not in supported_versions:
            raise ValueError(
                f"version '{version}' is not supported, it is not in {supported_versions}")
        self.version = version
        self.p_points_pu = p_points_pu[version]
        self.min_q_pu, self.max_q_pu = min_q_pu, max_q_pu
        self.q_max_under_p_point = kwargs.get("q_max_under_p_point", 0.)

        self.linear_factor_ind = (self.min_q_pu + 0.1) / (
            self.p_points_pu[1] - self.p_points_pu[0])
        self.linear_factor_cap = (self.max_q_pu - 0.1) / (
            self.p_points_pu[1] - self.p_points_pu[0])

    def in_area(self, p_pu, q_pu, vm_pu=None):
        is_in_area = np.ones(len(p_pu), dtype=bool)
        is_in_area[(p_pu < self.p_points_pu[0]) & ((q_pu < -0.05) | (q_pu > self.q_max_under_p_point))] = False
        if all(~is_in_area):
            return is_in_area
        is_in_area[(p_pu > self.p_points_pu[1]) & ((q_pu < self.min_q_pu) | (q_pu > self.max_q_pu))] = False
        if all(~is_in_area):
            return is_in_area
        is_in_area[(q_pu < -0.1-(p_pu-self.p_points_pu[0])*self.linear_factor_ind) | \
            (q_pu > 0.1+(p_pu-self.p_points_pu[0])*self.linear_factor_cap)] = False
        return is_in_area

    def q_flexibility(self, p_pu, vm_pu=None):
        q_flex = np.c_[[self.min_q_pu]*len(p_pu), [self.max_q_pu]*len(p_pu)]

        part = p_pu < self.p_points_pu[1]
        q_flex[part] = np.c_[-0.1+(p_pu[part]-self.p_points_pu[0])*self.linear_factor_ind,
                             0.1+(p_pu[part]-self.p_points_pu[0])*self.linear_factor_cap]

        part = p_pu < self.p_points_pu[0]
        q_flex[part] = np.c_[[-0.05]*sum(part), [self.q_max_under_p_point]*sum(part)]

        return q_flex


class QVArea4120(BaseArea):
    """ This class models the QV area of flexible Q for high-voltage power plants according to
    VDE AR-N-4120.
    It is used to be combined with active power dependencies in PQVArea4120V1, PQVArea4120V2, or
    PQVArea4120V3
    """
    def __init__(self, min_q_pu, max_q_pu):
        self.min_q_pu, self.max_q_pu = min_q_pu, max_q_pu

        self.max_vm_pu = 127.0/110
        self.min_vm_pu = 96.0/110
        self.delta_vm_pu = 7.0/110
        self.linear_factor = (self.max_q_pu - self.min_q_pu) / self.delta_vm_pu

    def q_flexibility(self, p_pu, vm_pu):

        # part = vm_pu > self.max_vm_pu
        len_ = len(vm_pu)
        min_max_q = np.c_[[self.min_q_pu]*len_, [self.min_q_pu]*len_]

        part = vm_pu < self.min_vm_pu
        len_ = sum(part)
        min_max_q[part] = np.c_[[self.max_q_pu]*len_, [self.max_q_pu]*len_]

        part = (self.min_vm_pu < vm_pu) & (vm_pu <= self.min_vm_pu + self.delta_vm_pu)
        min_max_q[part] = np.c_[self.max_q_pu - self.linear_factor * (vm_pu[part]-self.min_vm_pu),
                                [self.max_q_pu]*sum(part)]

        part = (self.min_vm_pu+self.delta_vm_pu < vm_pu) & (vm_pu <= self.max_vm_pu-self.delta_vm_pu)
        min_max_q[part] = np.repeat(np.array([[self.min_q_pu, self.max_q_pu]]), sum(part), axis=0)

        part = (self.max_vm_pu-self.delta_vm_pu < vm_pu) & (vm_pu <= self.max_vm_pu)
        min_max_q[part] = np.c_[[self.min_q_pu]*sum(part),
                                self.min_q_pu + self.linear_factor * (self.max_vm_pu-vm_pu[part])]

        return min_max_q


class PQVArea4120Base(BasePQVArea):
    """ This is the base class for the three variants of VDE AR-N-4120. This class is not for direct
    application by the user.
    """
    def __init__(self, min_q_pu, max_q_pu, version=2018, raise_merge_overlap=True):
        super().__init__(raise_merge_overlap=raise_merge_overlap)
        self.pq_area = PQArea4120(min_q_pu, max_q_pu, version=version)
        self.qv_area = QVArea4120(min_q_pu, max_q_pu)

class PQVArea4120V1(PQVArea4120Base):
    """ This class models the PQV area of flexible Q for high-voltage plants according to variant 1
    of VDE AR-N-4120.
    """
    def __init__(self, version=2018, raise_merge_overlap=True):
        super().__init__(-0.227902, 0.484322, version=version,
                         raise_merge_overlap=raise_merge_overlap)

class PQVArea4120V2(PQVArea4120Base):
    """ This class models the PQV area of flexible Q for high-voltage plants according to variant 2
    of VDE AR-N-4120.
    """
    def __init__(self, version=2018, raise_merge_overlap=True):
        super().__init__(-0.328684, 0.410775, version=version,
                         raise_merge_overlap=raise_merge_overlap)


class PQVArea4120V3(PQVArea4120Base):
    """ This class models the PQV area of flexible Q for high-voltage plants according to variant 3
    of VDE AR-N-4120.
    """
    def __init__(self, version=2018, raise_merge_overlap=True):
        super().__init__(-0.410775, 0.328684, version=version,
                         raise_merge_overlap=raise_merge_overlap)


""" EHV DERs: """

class PQArea4130(PQArea4120):
    """ This class models the PQ area of flexible Q for extra high-voltage plants according to
    VDE AR-N-4130. It is used to be combined with Voltage dependencies in PQVArea4130V1,
    PQVArea4130V2 and PQVArea4130V3.
    """
    def __init__(self, min_q_pu, max_q_pu):
        super().__init__(min_q_pu, max_q_pu, version=2018, q_max_under_p_point=0.05)


class QVArea4130(QVArea4120):
    """ This class models the QV area of flexible Q for extra high voltage power plants according to
    VDE AR-N-4130.
    It is used to be combined with active power dependencies in PQVArea4130V1, PQVArea4130V2, or
    PQVArea4130V3
    """
    def __init__(self, min_q_pu, max_q_pu, vn_kv, variant):
        self.min_q_pu, self.max_q_pu, self.vn_kv, self.variant = min_q_pu, max_q_pu, vn_kv, variant
        epsilon = 1e-3
        if not np.any(np.isclose(np.array([380, 220]), vn_kv)):
            raise ValueError(f"QVArea4130 is defined for 380kV and 220kV only, not for {vn_kv} kV.")
        if variant == 1:
            if vn_kv == 380:
                self.min_vm_points_pu = np.array([350-epsilon, 350, 380, 400]) / vn_kv
                self.max_vm_points_pu = np.array([420, 440]) / vn_kv
            elif vn_kv == 220:
                self.min_vm_points_pu = np.array([193-epsilon, 193, 220, 233.5]) / vn_kv
                self.max_vm_points_pu = np.array([245, 253]) / vn_kv
            self.min_q_points_pu = np.array([
                self.max_q_pu, self.max_q_pu*np.sin(np.arccos(0.95))/np.sin(np.arccos(0.9)), 0,
                                                    self.min_q_pu])
            self.max_q_points_pu = np.array([self.max_q_pu, self.min_q_pu])
        elif variant == 2:
            if vn_kv == 380:
                self.min_vm_points_pu = np.array([350-epsilon, 350, 380, 410]) / vn_kv
                self.max_vm_points_pu = np.array([420, 440, 440+epsilon]) / vn_kv
            elif vn_kv == 220:
                self.min_vm_points_pu = np.array([193-epsilon, 193, 220, 240]) / vn_kv
                self.max_vm_points_pu = np.array([245, 253, 253+epsilon]) / vn_kv
            self.min_q_points_pu = np.array([
                self.max_q, self.max_q*np.sin(np.arccos(0.95))/np.sin(np.arccos(0.925)), 0,
                                              self.min_q])
            self.max_q_points_pu = np.array([self.max_q, 0, self.min_q])
        elif variant == 3:
            if vn_kv == 380:
                self.min_vm_points_pu = np.array([350, 380]) / vn_kv
                self.max_vm_points_pu = np.array([420, 440, 440+epsilon]) / vn_kv
            elif vn_kv == 220:
                self.min_vm_points_pu = np.array([193, 220]) / vn_kv
                self.max_vm_points_pu = np.array([245, 253, 253+epsilon]) / vn_kv
            self.min_q_points_pu = np.array([self.max_q_pu, self.min_q_pu])
            self.max_q_points_pu = np.array([self.max_q_pu, 0, self.min_q_pu])
        else:
            raise ValueError(
                f"QVArea4130 is defined for variants 1, 2, and 3 but not for {self.variant}.")

    def q_flexibility(self, p_pu, vm_pu):
        return np.c_[np.interp(vm_pu, self.min_vm_points_pu, self.min_q_points_pu),
                     np.interp(vm_pu, self.max_vm_points_pu, self.max_q_points_pu)]


class PQVArea4130Base(BasePQVArea):
    """ This is the base class for the three variants of VDE AR-N-4130. This class is not for direct
    application by the user.
    """
    def __init__(self, min_q_pu, max_q_pu, variant, vn_kv=380, raise_merge_overlap=True):
        super().__init__(raise_merge_overlap=raise_merge_overlap)
        self.pq_area = PQArea4130(min_q_pu, max_q_pu)
        self.qv_area = QVArea4130(min_q_pu, max_q_pu, vn_kv=vn_kv, variant=variant)


class PQVArea4130V1(PQVArea4130Base):
    """ This class models the PQV area of flexible Q for extra high voltage power plants according
    to variant 1 of VDE AR-N-4130.
    """
    def __init__(self, vn_kv=380, raise_merge_overlap=True):
        super().__init__(-0.227902, 0.484322, 1, vn_kv=vn_kv, raise_merge_overlap=raise_merge_overlap)

class PQVArea4130V2(PQVArea4130Base):
    """ This class models the PQV area of flexible Q for extra high voltage power plants according
    to variant 2 of VDE AR-N-4130.
    """
    def __init__(self, vn_kv=380, raise_merge_overlap=True):
        super().__init__(-0.328684, 0.410775, 2, vn_kv=vn_kv, raise_merge_overlap=raise_merge_overlap)


class PQVArea4130V3(PQVArea4130Base):
    """ This class models the PQV area of flexible Q for extra high voltage power plants according
    to variant 3 of VDE AR-N-4130.
    """
    def __init__(self, vn_kv=380, raise_merge_overlap=True):
        super().__init__(-0.410775, 0.328684, 3, vn_kv=vn_kv, raise_merge_overlap=raise_merge_overlap)


""" MV DERs: """


class PQArea4110(PQAreaPOLYGON):
    def __init__(self):
        p_points_pu = (-1e-7,  0.05, 0.05       , 1.       , 1.      , 0.05      , 0.05, 1e-7, -1e-7)
        q_points_pu = (-1e-7, -1e-7, -0.01961505, -0.484322, 0.484322, 0.01961505, 0.  , 1e-7, -1e-7)
        super().__init__(p_points_pu, q_points_pu)


class QVArea4110(QVAreaPOLYGON):
    """
    This class models the QV area of flexible Q for medium-voltage plants according to
    VDE AR-N-4110.
    """
    def __init__(self):
        q_points_pu =  (0. , -0.484322, -0.484322, 0. , 0.484322, 0.484322, 0. )
        vm_points_pu = (0.9, 0.95     , 1.1      , 1.1, 1.05    , 0.9     , 0.9)
        super().__init__(q_points_pu, vm_points_pu)


class PQVArea4110(BasePQVArea):
    def __init__(self, raise_merge_overlap=True):
        super().__init__(raise_merge_overlap=raise_merge_overlap)
        self.pq_area = PQArea4110()
        self.qv_area = QVArea4110()

""" LV DERs: """


class PQArea4105(PQAreaPOLYGON):
    """
    This class models the PQ area of flexible Q for low-voltage plants according to VDE AR-N-4105.
    Plants with S_E,max <= 4.6 kVA should apply variant 1 while S_E,max > 4.6 kVA should apply
    variant 2.

    Note: To be in line with the other VDE AR N standard (meaning that all values are relative to
    p_{b,install}), the free area between the minimial reactive power requirements and the maximum
    apparent power of the inverter is not modeled as 'in_area'.
    """
    def __init__(self, variant):
        if variant == 1:
            # commented due to the note of the docstring
            # p_points_pu = (0., 0.95, 0.95, 0.)
            # q_points_pu = (0., -0.312, 0.312, 0.)
            p_points_pu = (0., 1, 1, 0.)
            q_points_pu = (0., -0.328684, 0.328684, 0.)
        elif variant == 2:
            # commented due to the note of the docstring
            # p_points_pu = (0., 0.9, 0.9, 0.)
            # q_points_pu = (0., -0.36, 0.36, 0.)
            p_points_pu = (0., 1, 1, 0.)
            q_points_pu = (0., -0.484322, 0.484322, 0.)
        else:
            raise ValueError(f"{variant=}")
        super().__init__(p_points_pu, q_points_pu)


class QVArea4105(QVAreaPOLYGON):
    """ This class models the QV area of flexible Q for low-voltage plants according to
    VDE AR-N-4105. Plants with S_E,max <= 4.6 kVA should apply variant 1 while S_E,max > 4.6 kVA
    should apply variant 2.
    """
    def __init__(self, variant):
        if variant == 1:
            q_points_pu =  (0. , -0.328684, -0.328684, 0. , 0.328684, 0.328684, 0. )
            vm_points_pu = (0.9, 0.95     , 1.1      , 1.1, 1.05    , 0.9     , 0.9)
        elif variant == 2:
            q_points_pu =  (0. , -0.484322, -0.484322, 0. , 0.484322, 0.484322, 0. )
            vm_points_pu = (0.9, 0.95     , 1.1      , 1.1, 1.05    , 0.9     , 0.9)
        else:
            raise ValueError(f"{variant=}")
        super().__init__(q_points_pu, vm_points_pu)


class PQVArea4105(BasePQVArea):
    """ This class models the PQV area of flexible Q for low-voltage plants according to
    VDE AR-N-4105. Plants with S_E,max <= 4.6 kVA should apply variant 1 while S_E,max > 4.6 kVA
    should apply variant 2.
    """
    def __init__(self, variant, raise_merge_overlap=True):
        super().__init__(raise_merge_overlap=raise_merge_overlap)
        self.pq_area = PQArea4105(variant)
        self.qv_area = QVArea4105(variant)


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

import numpy as np

from pandapower.auxiliary import read_from_net, write_to_net
from pandapower.control.controller.trafo_control import TrafoController


class DiscreteTapControl(TrafoController):
    """
    Trafo Controller with local tap changer voltage control.

    INPUT:
        **net** (attrdict) - Pandapower struct

        **element_index** (int) - ID of the trafo that is controlled

        **vm_lower_pu** (float) - Lower voltage limit in pu

        **vm_upper_pu** (float) - Upper voltage limit in pu

    OPTIONAL:

        **side** (string, "lv") - Side of the transformer where the voltage is controlled (hv or lv)

        **element** (string, "trafo") - Trafo type ("trafo" or "trafo3w")

        **tol** (float, 0.001) - Voltage tolerance band at bus in Percent (default: 1% = 0.01pu)

        **in_service** (bool, True) - Indicates if the controller is currently in_service

        **drop_same_existing_ctrl** (bool, False) - Indicates if already existing controllers of the same type and with the same matching parameters (e.g. at same element) should be dropped
    """

    def __init__(self, net, element_index, vm_lower_pu, vm_upper_pu, side="lv", element="trafo",
                 tol=1e-3, in_service=True, hunting_limit=None, level=0, order=0, drop_same_existing_ctrl=False,
                 matching_params=None, **kwargs):
        if matching_params is None:
            matching_params = {"element_index": element_index, 'element': element}
        super().__init__(net, element_index, side, tol=tol, in_service=in_service, level=level,
                         order=order, element=element,
                         drop_same_existing_ctrl=drop_same_existing_ctrl,
                         matching_params=matching_params,
                         **kwargs)

        self.vm_lower_pu = vm_lower_pu
        self.vm_upper_pu = vm_upper_pu

        self.vm_delta_pu = self.tap_step_percent / 100. * .5 + self.tol
        self.vm_set_pu = kwargs.get("vm_set_pu")
        self.hunting_limit = hunting_limit
        self._hunting_taps = np.array([], dtype=np.float64)

    @classmethod
    def from_tap_step_percent(cls, net, element_index, vm_set_pu, side="lv", element="trafo",
                              tol=1e-3, in_service=True, hunting_limit=None, level=0, order=0,
                              drop_same_existing_ctrl=False, matching_params=None, **kwargs):
        """
        Alternative mode of the controller, which uses a set point for voltage and the value of net.trafo.tap_step_percent to calculate
        vm_upper_pu and vm_lower_pu. To this end, the parameter vm_set_pu should be provided, instead of vm_lower_pu and vm_upper_pu.
        To use this mode of the controller, the controller can be initialized as following:

        >>> c = DiscreteTapControl.from_tap_step_percent(net, element_index, vm_set_pu)

        INPUT:
            **net** (attrdict) - Pandapower struct

            **element_index** (int) - ID of the trafo that is controlled

            **vm_set_pu** (float) - Voltage setpoint in pu
        """
        self = cls(net, element_index=element_index, vm_lower_pu=None, vm_upper_pu=None, side=side,
                   element=element, tol=tol,
                   in_service=in_service, hunting_limit=hunting_limit, level=level, order=order,
                   drop_same_existing_ctrl=drop_same_existing_ctrl,
                   matching_params=matching_params, vm_set_pu=vm_set_pu, **kwargs)
        return self

    @property
    def vm_set_pu(self):
        return self._vm_set_pu

    @vm_set_pu.setter
    def vm_set_pu(self, value):
        self._vm_set_pu = value
        if value is None:
            return
        self.vm_lower_pu = value - self.vm_delta_pu
        self.vm_upper_pu = value + self.vm_delta_pu

    def initialize_control(self, net):
        super().initialize_control(net)
        if hasattr(self, 'vm_set_pu') and self.vm_set_pu is not None:
            self.vm_delta_pu = self.tap_step_percent / 100. * .5 + self.tol
        if hasattr(self.element_index, "__iter__"):
            self._hunting_taps = np.full(shape=len(self.element_index), fill_value=np.nan,
                                         dtype=np.float64)
        else:
            self._hunting_taps = np.nan

    def control_step(self, net):
        """
        Implements one step of the Discrete controller, always stepping only one tap position up or down
        """
        if self.nothing_to_do(net):
            return

        vm_pu = read_from_net(net, "res_bus", self.trafobus, "vm_pu", self._read_write_flag)
        self.tap_pos = read_from_net(
            net, self.element, self.element_index, "tap_pos", self._read_write_flag)

        increment = np.where(
            self.tap_side_coeff * self.tap_sign == 1,
            np.where(np.logical_and(vm_pu < self.vm_lower_pu, self.tap_pos > self.tap_min), -1,
                     np.where(np.logical_and(vm_pu > self.vm_upper_pu, self.tap_pos < self.tap_max), 1, 0)),
            np.where(np.logical_and(vm_pu < self.vm_lower_pu, self.tap_pos < self.tap_max), 1,
                     np.where(np.logical_and(vm_pu > self.vm_upper_pu, self.tap_pos > self.tap_min), -1, 0)))

        self.tap_pos += increment

        self._hunting_taps = np.vstack([self._hunting_taps, self.tap_pos])
        if self.hunting_limit is not None and self._hunting_taps.shape[0] > self.hunting_limit:
            self._hunting_taps = self._hunting_taps[1:, :]

        # WRITE TO NET
        write_to_net(net, self.element, self.element_index, 'tap_pos',
                     self.tap_pos, self._read_write_flag)

    def is_converged(self, net):
        """
        Checks if the voltage is within the desired voltage band, then returns True
        """
        if self.nothing_to_do(net):
            return True

        vm_pu = read_from_net(net, "res_bus", self.trafobus, "vm_pu", self._read_write_flag)
        # this is possible in case the trafo is set out of service by the connectivity check
        is_nan = np.isnan(vm_pu)
        self.tap_pos = read_from_net(
            net, self.element, self.element_index, "tap_pos", self._read_write_flag)

        reached_limit = np.where(self.tap_side_coeff * self.tap_sign == 1,
                                 (vm_pu < self.vm_lower_pu) & (self.tap_pos == self.tap_min) |
                                 (vm_pu > self.vm_upper_pu) & (self.tap_pos == self.tap_max),
                                 (vm_pu < self.vm_lower_pu) & (self.tap_pos == self.tap_max) |
                                 (vm_pu > self.vm_upper_pu) & (self.tap_pos == self.tap_min))

        converged = np.logical_or(reached_limit, np.logical_and(self.vm_lower_pu < vm_pu, vm_pu < self.vm_upper_pu))

        return np.all(np.logical_or(converged, is_nan))



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.


import numpy as np
import pandas as pd
from pandapower.auxiliary import _sum_by_group, I_from_SV_elementwise, sequence_to_phase, S_from_VI_elementwise
from pandapower.pypower.idx_brch import F_BUS, T_BUS, PF, QF, PT, QT, BR_R
from pandapower.pypower.idx_brch_dc import DC_IF, DC_IT, DC_F_BUS, DC_T_BUS, DC_PF, DC_PT, DC_BR_R, DC_TDPF
from pandapower.pypower.idx_brch_tdpf import TDPF
from pandapower.pypower.idx_bus import BASE_KV, VM, VA
from pandapower.pypower.idx_bus_dc import DC_BASE_KV, DC_VM
from pandapower.pypower.idx_tcsc import TCSC_THYRISTOR_FIRING_ANGLE, TCSC_X_PU, TCSC_PF, TCSC_PT, TCSC_QF, TCSC_QT, \
    TCSC_IF, TCSC_IT


def _get_branch_results(net, ppc, bus_lookup_aranged, pq_buses, suffix=None):
    """
    Extract the branch results and writes it in the Dataframe net.res_line,  net.res_trafo, net.res_trafo3w...

    INPUT:

        **net** - pandapower net

        **ppc** (np.array) - ppc structure

        **bus_lookup_aranged** - consecutive aranged bus lookup

        **pq_buses** - the PQ type buses in ppc

    OPTIONAL:

        **suffix** (str, None) - a suffix like "_sc" to write to res_line_sc for example (short circuit)

    """

    i_ft, s_ft = _get_branch_flows(ppc)

    _get_line_results(net, ppc, i_ft, suffix=suffix)
    _get_line_dc_results(net, ppc)
    _get_trafo_results(net, ppc, s_ft, i_ft, suffix=suffix)
    _get_trafo3w_results(net, ppc, s_ft, i_ft, suffix=suffix)
    _get_impedance_results(net, ppc, i_ft, suffix=suffix)
    _get_xward_branch_results(net, ppc, bus_lookup_aranged, pq_buses, suffix=suffix)
    _get_switch_results(net, ppc, i_ft, suffix=suffix)
    _get_tcsc_results(net, ppc, suffix=suffix)


def _get_branch_results_3ph(net, ppc0, ppc1, ppc2, bus_lookup_aranged, pq_buses):
    """
    Extract the bus results and writes it in the Dataframe net.res_line and net.res_trafo.

    INPUT:

        **results** - the result of runpf loadflow calculation

        **p** - the dict to dump the "res_line" and "res_trafo" Dataframe

    """
    I012_f, S012_f, V012_f, I012_t, S012_t, V012_t = _get_branch_flows_3ph(ppc0, ppc1, ppc2)
    _get_line_results_3ph(net, ppc0, ppc1, ppc2, I012_f, V012_f, I012_t, V012_t)
    _get_trafo_results_3ph(net, ppc0, ppc1, ppc2, I012_f, V012_f, I012_t, V012_t)
    # _get_trafo3w_results(net, ppc, s_ft, i_ft)
    # _get_impedance_results(net, ppc, i_ft)
    # _get_xward_branch_results(net, ppc, bus_lookup_aranged, pq_buses)
    # _get_switch_results(net, i_ft)


def _get_branch_flows(ppc):
    br_idx = ppc["branch"][:, (F_BUS, T_BUS)].real.astype(np.int64)
    vm_ft = ppc["bus"][br_idx, VM] * ppc["bus"][br_idx, BASE_KV]
    s_ft = np.sqrt(ppc["branch"][:, (PF, PT)].real ** 2 +
                   ppc["branch"][:, (QF, QT)].real ** 2)
    i_ft = s_ft / vm_ft / np.sqrt(3)
    return i_ft, s_ft


def _get_branch_flows_3ph(ppc0, ppc1, ppc2):
    br_from_idx = ppc1["branch"][:, F_BUS].real.astype(np.int64)
    br_to_idx = ppc1["branch"][:, T_BUS].real.astype(np.int64)
    V012_f = np.array([(ppc["bus"][br_from_idx, VM] * ppc["bus"][br_from_idx, BASE_KV] *
                        np.exp(1j * np.deg2rad(ppc["bus"][br_from_idx, VA]))).flatten() for ppc in [ppc0, ppc1, ppc2]])
    V012_t = np.array([(ppc["bus"][br_to_idx, VM] * ppc["bus"][br_to_idx, BASE_KV] *
                        np.exp(1j * np.deg2rad(ppc["bus"][br_to_idx, VA]))).flatten() for ppc in [ppc0, ppc1, ppc2]])
    S012_f = np.array([((ppc["branch"][:, PF].real +
                         1j * ppc["branch"][:, QF].real))
                       for ppc in [ppc0, ppc1, ppc2]])
    S012_t = np.array([((ppc["branch"][:, PT].real +
                         1j * ppc["branch"][:, QT].real))
                       for ppc in [ppc0, ppc1, ppc2]])
    I012_f = I_from_SV_elementwise(S012_f, V012_f / np.sqrt(3))
    I012_t = I_from_SV_elementwise(S012_t, V012_t / np.sqrt(3))

    return I012_f, S012_f, V012_f, I012_t, S012_t, V012_t


def _get_line_results(net, ppc, i_ft, suffix=None):
    # create res_line_vals which are written to the pandas dataframe
    if "line" not in net._pd2ppc_lookups["branch"]:
        return
    ac = net["_options"]["ac"]

    f, t = net._pd2ppc_lookups["branch"]["line"]
    pf_mw = ppc["branch"][f:t, PF].real
    q_from_mvar = ppc["branch"][f:t, QF].real
    p_from_mw = pf_mw

    pt_mw = ppc["branch"][f:t, PT].real
    q_to_mvar = ppc["branch"][f:t, QT].real
    p_to_mw = pt_mw

    if ac:
        pl_mw = pf_mw + pt_mw
        ql_mvar = q_from_mvar + q_to_mvar
    else:
        pl_mw = np.zeros_like(pf_mw)
        ql_mvar = np.zeros_like(q_from_mvar)

    with np.errstate(invalid='ignore'):
        i_ka = np.max(i_ft[f:t], axis=1)
    i_from_ka = i_ft[f:t][:, 0]
    i_to_ka = i_ft[f:t][:, 1]
    line_df = net["line"]
    i_max = line_df["max_i_ka"].values * line_df["df"].values * line_df["parallel"].values
    from_bus = ppc["branch"][f:t, F_BUS].real.astype(np.int64)
    to_bus = ppc["branch"][f:t, T_BUS].real.astype(np.int64)

    # write to line
    res_line_df = net["res_line"] if suffix is None else net["res_line%s" % suffix]

    res_line_df["p_from_mw"].values[:] = p_from_mw
    res_line_df["q_from_mvar"].values[:] = q_from_mvar
    res_line_df["p_to_mw"].values[:] = p_to_mw
    res_line_df["q_to_mvar"].values[:] = q_to_mvar
    res_line_df["pl_mw"].values[:] = pl_mw
    res_line_df["ql_mvar"].values[:] = ql_mvar
    res_line_df["i_from_ka"].values[:] = i_from_ka
    res_line_df["i_to_ka"].values[:] = i_to_ka
    res_line_df["i_ka"].values[:] = i_ka
    res_line_df["vm_from_pu"].values[:] = ppc["bus"][from_bus, VM]
    res_line_df["va_from_degree"].values[:] = ppc["bus"][from_bus, VA]
    res_line_df["vm_to_pu"].values[:] = ppc["bus"][to_bus, VM]
    res_line_df["va_to_degree"].values[:] = ppc["bus"][to_bus, VA]
    loading = np.full_like(i_ka, fill_value=np.inf, dtype=np.float64)
    np.divide(i_ka, i_max, where=i_max != 0, out=loading, dtype=np.float64)
    res_line_df["loading_percent"].values[:] = loading * 100

    # if consider_line_temperature, add resulting r_ohm_per_km to net.res_line
    if net["_options"]["consider_line_temperature"] or net["_options"].get("tdpf", False):
        base_kv = ppc["bus"][from_bus, BASE_KV]
        baseR = np.square(base_kv) / net.sn_mva
        length_km = line_df.length_km.values
        parallel = line_df.parallel.values
        res_line_df["r_ohm_per_km"] = ppc["branch"][f:t, BR_R].real / length_km * baseR * parallel

        if net["_options"].get("tdpf", False):
            tdpf_lines = ppc["internal"]['branch_is'][f:t] & np.nan_to_num(ppc['branch'][f:t, TDPF]).real.astype(bool)
            res_line_df.loc[tdpf_lines, "r_theta_kelvin_per_mw"] = ppc["internal"]["r_theta_kelvin_per_mw"]
            no_tdpf_t = line_df.loc[~tdpf_lines].get("temperature_degree_celsius", default=20.)
            res_line_df.loc[tdpf_lines, "temperature_degree_celsius"] = ppc["internal"]["T"]
            res_line_df.loc[~tdpf_lines, "temperature_degree_celsius"] = no_tdpf_t


def _get_line_dc_results(net, ppc):
    # create res_line_vals which are written to the pandas dataframe
    if "line_dc" not in net._pd2ppc_lookups["branch_dc"]:
        return
    ac = net["_options"]["ac"]
    if not ac:
        return

    f, t = net._pd2ppc_lookups["branch_dc"]["line_dc"]
    from_bus = ppc["branch_dc"][f:t, DC_F_BUS].real.astype(np.int64)
    to_bus = ppc["branch_dc"][f:t, DC_T_BUS].real.astype(np.int64)
    base_v_kv = ppc["bus_dc"][from_bus, DC_BASE_KV]
    base_i_ka = ppc["baseMVA"] / base_v_kv

    p_from_mw = ppc["branch_dc"][f:t, DC_PF].real
    i_from_ka = ppc["branch_dc"][f:t, DC_IF].real * base_i_ka

    p_to_mw = ppc["branch_dc"][f:t, DC_PT].real
    i_to_ka = ppc["branch_dc"][f:t, DC_IT].real * base_i_ka

    i_ka = np.maximum(i_from_ka, i_to_ka)

    line_df = net["line_dc"]
    i_max = line_df["max_i_ka"].values * line_df["df"].values * line_df["parallel"].values

    # write to line
    res_line_df = net["res_line_dc"]

    res_line_df["p_from_mw"].values[:] = p_from_mw
    res_line_df["p_to_mw"].values[:] = p_to_mw
    res_line_df["pl_mw"].values[:] = p_from_mw + p_to_mw
    res_line_df["i_from_ka"].values[:] = i_from_ka
    res_line_df["i_to_ka"].values[:] = i_to_ka
    res_line_df["i_ka"].values[:] = i_ka
    res_line_df["vm_from_pu"].values[:] = ppc["bus_dc"][from_bus, DC_VM]
    res_line_df["vm_to_pu"].values[:] = ppc["bus_dc"][to_bus, DC_VM]
    loading = np.full_like(i_ka, fill_value=np.inf, dtype=np.float64)
    np.divide(i_ka, i_max, where=i_max != 0, out=loading, dtype=np.float64)
    res_line_df["loading_percent"].values[:] = loading * 100

    # if consider_line_temperature, add resulting r_ohm_per_km to net.res_line
    if net["_options"]["consider_line_temperature"] or net["_options"].get("tdpf", False):
        baseR = np.square(base_v_kv) / net.sn_mva
        length_km = line_df.length_km.values
        parallel = line_df.parallel.values
        res_line_df["r_ohm_per_km"] = ppc["branch_dc"][f:t, DC_BR_R].real / length_km * baseR * parallel

        if net["_options"].get("tdpf", False):
            tdpf_lines = ppc["internal"]['branch_dc_is'][f:t] & np.nan_to_num(ppc['branch_dc'][f:t, DC_TDPF]).real.astype(bool)
            # res_line_df.loc[tdpf_lines, "r_theta_kelvin_per_mw"] = ppc["internal"]["r_theta_kelvin_per_mw"]  # todo for dc
            no_tdpf_t = line_df.loc[~tdpf_lines].get("temperature_degree_celsius", default=20.)
            # res_line_df.loc[tdpf_lines, "temperature_degree_celsius"] = ppc["internal"]["T"]  # todo for DC
            res_line_df.loc[~tdpf_lines, "temperature_degree_celsius"] = no_tdpf_t


def _get_line_results_3ph(net, ppc0, ppc1, ppc2, I012_f, V012_f, I012_t, V012_t):
    # create res_line_vals which are written to the pandas dataframe
    ac = net["_options"]["ac"]

    if not "line" in net._pd2ppc_lookups["branch"]:
        return

    f, t = net._pd2ppc_lookups["branch"]["line"]
    I012_from_ka = I012_f[:, f:t]
    I012_to_ka = I012_t[:, f:t]
    line_df = net["line"]
    i_max_phase = line_df["max_i_ka"].values * line_df["df"].values * line_df["parallel"].values

    Vabc_f, Vabc_t, Iabc_f, Iabc_t = [sequence_to_phase(X012) for X012 in
                                      [V012_f[:, f:t], V012_t[:, f:t], I012_f[:, f:t], I012_t[:, f:t]]]
    Sabc_f, Sabc_t = [S_from_VI_elementwise(*Xabc_tup) / np.sqrt(3) for Xabc_tup in
                      [(Vabc_f, Iabc_f), (Vabc_t, Iabc_t)]]
    # Todo: Check why the sqrt(3) is necessary in the previous line as opposed to _get_line_results()
    Pabcf_mw = Sabc_f.real
    Qabcf_mvar = Sabc_f.imag
    Pabct_mw = Sabc_t.real
    Qabct_mvar = Sabc_t.imag
    if ac:
        Pabcl_mw = Pabcf_mw + Pabct_mw
        Qabcl_mvar = Qabcf_mvar + Qabct_mvar
    else:
        Pabcl_mw = np.zeros_like(Pabcf_mw)
        Qabcl_mvar = np.zeros_like(Qabct_mvar)

    # getting complex values of the sequence current line
    Iabc_f_ka_complex = sequence_to_phase(I012_from_ka)
    Iabc_t_ka_complex = sequence_to_phase(I012_to_ka)

    Iabc_f_ka = np.abs(Iabc_f_ka_complex)
    Iabc_t_ka = np.abs(Iabc_t_ka_complex)
    Iabc_ka = np.maximum.reduce([Iabc_t_ka, Iabc_f_ka])

    In_f_ka_complex = Iabc_f_ka_complex.sum(axis=0)
    In_f_ka = np.abs(In_f_ka_complex)
#    In_f_ia_n_degree = np.angle(In_f_ka_complex).flatten()*180/np.pi
    In_t_ka_complex = Iabc_t_ka_complex.sum(axis=0)
#    In_t_ia_n_degree = np.angle(In_t_ka_complex).flatten()*180/np.pi
    In_t_ka = np.abs(In_t_ka_complex)
    In_ka = np.maximum.reduce([In_t_ka, In_f_ka])

    # write to line
    net["res_line_3ph"]["p_a_from_mw"] = Pabcf_mw[0, :].flatten()
    net["res_line_3ph"]["p_b_from_mw"] = Pabcf_mw[1, :].flatten()
    net["res_line_3ph"]["p_c_from_mw"] = Pabcf_mw[2, :].flatten()
    net["res_line_3ph"]["q_a_from_mvar"] = Qabcf_mvar[0, :].flatten()
    net["res_line_3ph"]["q_b_from_mvar"] = Qabcf_mvar[1, :].flatten()
    net["res_line_3ph"]["q_c_from_mvar"] = Qabcf_mvar[2, :].flatten()
    net["res_line_3ph"]["p_a_to_mw"] = Pabct_mw[0, :].flatten()
    net["res_line_3ph"]["p_b_to_mw"] = Pabct_mw[1, :].flatten()
    net["res_line_3ph"]["p_c_to_mw"] = Pabct_mw[2, :].flatten()
    net["res_line_3ph"]["q_a_to_mvar"] = Qabct_mvar[0, :].flatten()
    net["res_line_3ph"]["q_b_to_mvar"] = Qabct_mvar[1, :].flatten()
    net["res_line_3ph"]["q_c_to_mvar"] = Qabct_mvar[2, :].flatten()
    net["res_line_3ph"]["p_a_l_mw"] = Pabcl_mw[0, :].flatten()
    net["res_line_3ph"]["p_b_l_mw"] = Pabcl_mw[1, :].flatten()
    net["res_line_3ph"]["p_c_l_mw"] = Pabcl_mw[2, :].flatten()
    net["res_line_3ph"]["q_a_l_mvar"] = Qabcl_mvar[0, :].flatten()
    net["res_line_3ph"]["q_b_l_mvar"] = Qabcl_mvar[1, :].flatten()
    net["res_line_3ph"]["q_c_l_mvar"] = Qabcl_mvar[2, :].flatten()
    net["res_line_3ph"]["i_a_from_ka"] = Iabc_f_ka[0, :].flatten()
    net["res_line_3ph"]["i_b_from_ka"] = Iabc_f_ka[1, :].flatten()
    net["res_line_3ph"]["i_c_from_ka"] = Iabc_f_ka[2, :].flatten()
    net["res_line_3ph"]["i_a_to_ka"] = Iabc_t_ka[0, :].flatten()
    net["res_line_3ph"]["i_b_to_ka"] = Iabc_t_ka[1, :].flatten()
    net["res_line_3ph"]["i_c_to_ka"] = Iabc_t_ka[2, :].flatten()
    net["res_line_3ph"]["i_a_ka"] = Iabc_ka[0, :]
    net["res_line_3ph"]["i_b_ka"] = Iabc_ka[1, :]
    net["res_line_3ph"]["i_c_ka"] = Iabc_ka[2, :]
    net["res_line_3ph"]["i_n_from_ka"] = In_f_ka
    net["res_line_3ph"]["i_n_to_ka"] = In_t_ka
    net["res_line_3ph"]["i_n_ka"] = In_ka
    net["res_line_3ph"]["loading_a_percent"] = Iabc_ka[0, :] / i_max_phase * 100
    net["res_line_3ph"]["loading_b_percent"] = Iabc_ka[1, :] / i_max_phase * 100
    net["res_line_3ph"]["loading_c_percent"] = Iabc_ka[2, :] / i_max_phase * 100
    net["res_line_3ph"]["loading_percent"] = Iabc_ka.max(axis=0) / i_max_phase * 100
    net["res_line_3ph"].index = net["line"].index


def _get_trafo_results(net, ppc, s_ft, i_ft, suffix=None):
    if "trafo" not in net._pd2ppc_lookups["branch"]:
        return
    ac = net["_options"]["ac"]
    trafo_loading = net["_options"]["trafo_loading"]
    f, t = net._pd2ppc_lookups["branch"]["trafo"]
    p_hv_mw = ppc["branch"][f:t, PF].real
    p_lv_mw = ppc["branch"][f:t, PT].real

    if ac:
        q_hv_mvar = ppc["branch"][f:t, QF].real
        q_lv_mvar = ppc["branch"][f:t, QT].real
        pl_mw = p_hv_mw + p_lv_mw
        ql_mvar = q_hv_mvar + q_lv_mvar
    else:
        q_hv_mvar = np.zeros_like(p_hv_mw)
        q_lv_mvar = np.zeros_like(p_lv_mw)
        pl_mw = np.zeros_like(p_lv_mw)
        ql_mvar = np.zeros_like(p_lv_mw)

    i_hv_ka = i_ft[:, 0][f:t]
    i_lv_ka = i_ft[:, 1][f:t]
    if trafo_loading == "current":
        # calculate loading with rated current
        trafo_df = net["trafo"]
        vns = np.vstack([trafo_df["vn_hv_kv"].values, trafo_df["vn_lv_kv"].values]).T
        lds_trafo = i_ft[f:t] * vns * np.sqrt(3) / trafo_df["sn_mva"].values[:, np.newaxis] * 100.
        with np.errstate(invalid='ignore'):
            ld_trafo = np.max(lds_trafo, axis=1)
    elif trafo_loading == "power":
        # calculate loading with rated loading
        ld_trafo = np.max(s_ft[f:t] / net["trafo"]["sn_mva"].values[:, np.newaxis] * 100., axis=1)
    else:
        raise ValueError(
            "Unknown transformer loading parameter %s - choose 'current' or 'power'" % trafo_loading)
    if any(net["trafo"]["df"].values <= 0):
        raise UserWarning('Transformer rating factor df must be positive. Transformers with false '
                          'rating factors: %s' % net["trafo"].query('df<=0').index.tolist())

    loading_percent = ld_trafo / net["trafo"]["parallel"].values / net["trafo"]["df"].values

    hv_buses = ppc["branch"][f:t, F_BUS].real.astype(np.int64)
    lv_buses = ppc["branch"][f:t, T_BUS].real.astype(np.int64)

    # write results to trafo dataframe
    res_trafo_df = net["res_trafo"] if suffix is None else net["res_trafo%s" % suffix]

    res_trafo_df["p_hv_mw"].values[:] = p_hv_mw
    res_trafo_df["q_hv_mvar"].values[:] = q_hv_mvar
    res_trafo_df["p_lv_mw"].values[:] = p_lv_mw
    res_trafo_df["q_lv_mvar"].values[:] = q_lv_mvar
    res_trafo_df["pl_mw"].values[:] = pl_mw
    res_trafo_df["ql_mvar"].values[:] = ql_mvar
    res_trafo_df["i_hv_ka"].values[:] = i_hv_ka
    res_trafo_df["i_lv_ka"].values[:] = i_lv_ka
    res_trafo_df["vm_hv_pu"].values[:] = ppc["bus"][hv_buses, VM]
    res_trafo_df["va_hv_degree"].values[:] = ppc["bus"][hv_buses, VA]
    res_trafo_df["vm_lv_pu"].values[:] = ppc["bus"][lv_buses, VM]
    res_trafo_df["va_lv_degree"].values[:] = ppc["bus"][lv_buses, VA]
    res_trafo_df["loading_percent"].values[:] = loading_percent


def _get_trafo_results_3ph(net, ppc0, ppc1, ppc2, I012_f, V012_f, I012_t, V012_t):
    ac = net["_options"]["ac"]
    trafo_loading = net["_options"]["trafo_loading"]

    if not "trafo" in net._pd2ppc_lookups["branch"]:
        return
    f, t = net._pd2ppc_lookups["branch"]["trafo"]
    I012_hv_ka = I012_f[:, f:t]
    I012_lv_ka = I012_t[:, f:t]
    trafo_df = net["trafo"]

    Vabc_hv, Vabc_lv, Iabc_hv, Iabc_lv = [sequence_to_phase(X012) for X012 in
                                          [V012_f[:, f:t], V012_t[:, f:t], I012_f[:, f:t], I012_t[:, f:t]]]
    Sabc_hv, Sabc_lv = [S_from_VI_elementwise(*Xabc_tup) / np.sqrt(3) for Xabc_tup in
                        [(Vabc_hv, Iabc_hv), (Vabc_lv, Iabc_lv)]]
    # Todo: Check why the sqrt(3) is necessary in the previous line as opposed to _get_line_results()
    Pabc_hv_mw = Sabc_hv.real
    Qabc_hv_mvar = Sabc_hv.imag
    Pabc_lv_mw = Sabc_lv.real
    Qabc_lv_mvar = Sabc_lv.imag
    if ac:
        Pabcl_mw = Pabc_hv_mw + Pabc_lv_mw
        Qabcl_mvar = Qabc_hv_mvar + Qabc_lv_mvar
    else:
        Pabcl_mw = np.zeros_like(Pabc_hv_mw)
        Qabcl_mvar = np.zeros_like(Qabc_lv_mvar)

    Iabc_hv_ka = np.abs(sequence_to_phase(I012_hv_ka))
    Iabc_lv_ka = np.abs(sequence_to_phase(I012_lv_ka))

    # current calculation for trafo lv side for vector groups with zero seq. gap (Dyn, Yzn)
    # in this case, the currents of elemnts that go out from the trafo are summed and the sum applied to the trafo lv side
    gap_trafo_index = np.where(I012_lv_ka[0] == 0)[0]
    if len(gap_trafo_index > 0):
        for i_trafo in gap_trafo_index:
            Iabc_sum = [0, 0, 0]
            lv_bus = net.trafo.lv_bus.iat[i_trafo]
            V_bus_abc = np.array([[net.res_bus_3ph['vm_a_pu'][lv_bus] * net.bus['vn_kv'][lv_bus]],
                                  [net.res_bus_3ph['vm_b_pu'][lv_bus] * net.bus['vn_kv'][lv_bus]],
                                  [net.res_bus_3ph['vm_c_pu'][lv_bus] * net.bus['vn_kv'][lv_bus]]])

            # Branch Elements
            i_branch = np.concatenate((np.where(ppc0['branch'][:, F_BUS] == lv_bus)[0],
                                       np.where(ppc0['branch'][:, T_BUS] == lv_bus)[0]))
            i_branch = np.delete(i_branch, np.where(i_branch == i_trafo + f))  # delete the trafo itself from the list
            if len(i_branch > 0):
                I_branch_012 = I012_f[:, i_branch]
                I_branch_abc = sequence_to_phase(I_branch_012)
                for x in range(len(I_branch_abc[0])):
                    Iabc_sum += abs(I_branch_abc[:, x])

            # Loads
            load_index = net.asymmetric_load.index[net.asymmetric_load['bus'] == lv_bus]
            if len(load_index > 0):
                S_load_abc = abs(np.array([
                    np.array(net.res_asymmetric_load_3ph['p_a_mw'][load_index]
                             + (1j * net.res_asymmetric_load_3ph['q_a_mvar'][load_index])),
                    np.array(net.res_asymmetric_load_3ph['p_b_mw'][load_index]
                             + (1j * net.res_asymmetric_load_3ph['q_b_mvar'][load_index])),
                    np.array(net.res_asymmetric_load_3ph['p_c_mw'][load_index]
                             + (1j * net.res_asymmetric_load_3ph['q_c_mvar'][load_index]))]))
                I_load_abc = S_load_abc / (V_bus_abc / np.sqrt(3))
                for x in range(len(I_load_abc[0])):
                    Iabc_sum += I_load_abc[:, x]

            # Sgens
            sgen_bus_index = np.where(net.asymmetric_sgen['bus'] == lv_bus)[0]
            if len(sgen_bus_index > 0):
                S_sgen_abc = abs(np.array([
                    np.array(net.res_asymmetric_sgen_3ph['p_a_mw'][sgen_bus_index]
                             + (1j * net.res_asymmetric_sgen_3ph['q_a_mvar'][sgen_bus_index])),
                    np.array(net.res_asymmetric_sgen_3ph['p_b_mw'][sgen_bus_index]
                             + (1j * net.res_asymmetric_sgen_3ph['q_b_mvar'][sgen_bus_index])),
                    np.array(net.res_asymmetric_sgen_3ph['p_c_mw'][sgen_bus_index]
                             + (1j * net.res_asymmetric_sgen_3ph['q_c_mvar'][sgen_bus_index]))]))
                I_sgen_abc = S_sgen_abc / (V_bus_abc / np.sqrt(3))
                for x in range(len(I_sgen_abc[0])):
                    Iabc_sum -= I_sgen_abc[:, x]

            Iabc_lv_ka[:, i_trafo] = Iabc_sum

    # geting complex values of the sequence current
#    Iabc_hv_ka_complex = sequence_to_phase(I012_hv_ka)
#    Iabc_lv_ka_complex = sequence_to_phase(I012_lv_ka)
#
#    Iabc_hv_ka = np.abs(Iabc_hv_ka_complex)
#    Iabc_lv_ka = np.abs(Iabc_lv_ka_complex)
#
#    In_hv_ka_complex = Iabc_hv_ka_complex.sum(axis=0)
#    In_hv_ka = np.abs(In_hv_ka_complex)
#    In_hv_ia_n_degree = np.angle(In_hv_ka_complex).flatten()*180/np.pi
#    In_lv_ka_complex = Iabc_lv_ka_complex.sum(axis=0)
#    In_lv_ka = np.abs(In_lv_ka_complex)
#    In_lv_ia_n_degree = np.angle(In_lv_ka_complex).flatten()*180/np.pi

    if trafo_loading == "current":
        trafo_df = net["trafo"]
        vns = np.vstack([trafo_df["vn_hv_kv"].values, trafo_df["vn_lv_kv"].values]).T
        ld_trafo = np.maximum.reduce([np.asarray(Iabc_hv_ka) * vns[:, 0], np.asarray(Iabc_lv_ka) * vns[:, 1]])
        ld_trafo = ld_trafo * np.sqrt(3) / trafo_df["sn_mva"].values * 100.
    elif trafo_loading == "power":
        ld_trafo = np.maximum.reduce([np.abs(Sabc_hv), np.abs(Sabc_lv)])
        ld_trafo = ld_trafo / net["trafo"]["sn_mva"].values[:, np.newaxis] * 3. * 100.
    else:
        raise ValueError(
            "Unknown transformer loading parameter %s - choose 'current' or 'power'" % trafo_loading)
    if any(net["trafo"]["df"].values <= 0):
        raise UserWarning('Transformer rating factor df must be positive. Transformers with false '
                          'rating factors: %s' % net["trafo"].query('df<=0').index.tolist())
    loading_percent = ld_trafo / net["trafo"]["parallel"].values / net["trafo"]["df"].values

    # write results to trafo dataframe
    res_trafo_df = net["res_trafo_3ph"]
    res_trafo_df["p_a_hv_mw"] = Pabc_hv_mw[0, :].flatten()
    res_trafo_df["p_b_hv_mw"] = Pabc_hv_mw[1, :].flatten()
    res_trafo_df["p_c_hv_mw"] = Pabc_hv_mw[2, :].flatten()
    res_trafo_df["q_a_hv_mvar"] = Qabc_hv_mvar[0, :].flatten()
    res_trafo_df["q_b_hv_mvar"] = Qabc_hv_mvar[1, :].flatten()
    res_trafo_df["q_c_hv_mvar"] = Qabc_hv_mvar[2, :].flatten()
    res_trafo_df["p_a_lv_mw"] = Pabc_lv_mw[0, :].flatten()
    res_trafo_df["p_b_lv_mw"] = Pabc_lv_mw[1, :].flatten()
    res_trafo_df["p_c_lv_mw"] = Pabc_lv_mw[2, :].flatten()
    res_trafo_df["q_a_lv_mvar"] = Qabc_lv_mvar[0, :].flatten()
    res_trafo_df["q_b_lv_mvar"] = Qabc_lv_mvar[1, :].flatten()
    res_trafo_df["q_c_lv_mvar"] = Qabc_lv_mvar[2, :].flatten()
    res_trafo_df["p_a_l_mw"] = Pabcl_mw[0, :].flatten()
    res_trafo_df["p_b_l_mw"] = Pabcl_mw[1, :].flatten()
    res_trafo_df["p_c_l_mw"] = Pabcl_mw[2, :].flatten()
    res_trafo_df["q_a_l_mvar"] = Qabcl_mvar[0, :].flatten()
    res_trafo_df["q_b_l_mvar"] = Qabcl_mvar[1, :].flatten()
    res_trafo_df["q_c_l_mvar"] = Qabcl_mvar[2, :].flatten()
    res_trafo_df["i_a_hv_ka"] = Iabc_hv_ka[0, :].flatten()
    res_trafo_df["i_b_hv_ka"] = Iabc_hv_ka[1, :].flatten()
    res_trafo_df["i_c_hv_ka"] = Iabc_hv_ka[2, :].flatten()
#    res_trafo_df["i_n_hv_ka"] = In_hv_ka.flatten()
    res_trafo_df["i_a_lv_ka"] = Iabc_lv_ka[0, :].flatten()
    res_trafo_df["i_b_lv_ka"] = Iabc_lv_ka[1, :].flatten()
    res_trafo_df["i_c_lv_ka"] = Iabc_lv_ka[2, :].flatten()
#    res_trafo_df["i_n_lv_ka"] = In_lv_ka.flatten()
    res_trafo_df["loading_a_percent"] = loading_percent[0, :]
    res_trafo_df["loading_b_percent"] = loading_percent[1, :]
    res_trafo_df["loading_c_percent"] = loading_percent[2, :]
    res_trafo_df["loading_percent"] = loading_percent.max(axis=0)
    res_trafo_df.index = net["trafo"].index.values


def _get_trafo3w_lookups(net):
    f, t = net._pd2ppc_lookups["branch"]["trafo3w"]
    hv = int(f + (t - f) / 3)
    mv = int(f + 2 * (t - f) / 3)
    lv = t
    return f, hv, mv, lv


def _get_trafo3w_results(net, ppc, s_ft, i_ft, suffix=None):
    if "trafo3w" not in net._pd2ppc_lookups["branch"]:
        return

    trafo_loading = net["_options"]["trafo_loading"]
    ac = net["_options"]["ac"]
    f, hv, mv, lv = _get_trafo3w_lookups(net)

    phv_mw = ppc["branch"][f:hv, PF].real
    pmv_mw = ppc["branch"][hv:mv, PT].real
    plv_mw = ppc["branch"][mv:lv, PT].real

    p_hv_mw = phv_mw
    p_mv_mw = pmv_mw
    p_lv_mw = plv_mw

    if ac:
        q_hv_mvar = ppc["branch"][f:hv, QF].real
        q_mv_mvar = ppc["branch"][hv:mv, QT].real
        q_lv_mvar = ppc["branch"][mv:lv, QT].real
        pl_mw = phv_mw + pmv_mw + plv_mw
        ql_mvar = q_hv_mvar + q_mv_mvar + q_lv_mvar
    else:
        zeros_ = np.zeros_like(phv_mw)
        q_hv_mvar = zeros_
        q_mv_mvar = zeros_
        q_lv_mvar = zeros_
        pl_mw = zeros_
        ql_mvar = zeros_

    i_h = i_ft[:, 0][f:hv]
    i_m = i_ft[:, 1][hv:mv]
    i_l = i_ft[:, 1][mv:lv]

    t3 = net["trafo3w"]
    if trafo_loading == "current":
        ld_h = i_h * t3["vn_hv_kv"].values * np.sqrt(3) / t3["sn_hv_mva"].values * 100
        ld_m = i_m * t3["vn_mv_kv"].values * np.sqrt(3) / t3["sn_mv_mva"].values * 100
        ld_l = i_l * t3["vn_lv_kv"].values * np.sqrt(3) / t3["sn_lv_mva"].values * 100
        with np.errstate(invalid='ignore'):
            ld_trafo = np.max(np.vstack([ld_h, ld_m, ld_l]), axis=0)
    elif trafo_loading == "power":
        ld_h = s_ft[:, 0][f:hv] / t3["sn_hv_mva"].values * 100.
        ld_m = s_ft[:, 1][hv:mv] / t3["sn_mv_mva"].values * 100.
        ld_l = s_ft[:, 1][mv:lv] / t3["sn_lv_mva"].values * 100.
        ld_trafo = np.max(np.vstack([ld_h, ld_m, ld_l]), axis=0)
    else:
        raise ValueError(
            "Unknown transformer loading parameter %s - choose 'current' or 'power'" % trafo_loading)
    loading_percent = ld_trafo

    hv_buses = ppc["branch"][f:hv, F_BUS].real.astype(np.int64)
    aux_buses = ppc["branch"][f:hv, T_BUS].real.astype(np.int64)
    mv_buses = ppc["branch"][hv:mv, T_BUS].real.astype(np.int64)
    lv_buses = ppc["branch"][mv:lv, T_BUS].real.astype(np.int64)

    # write results to trafo3w dataframe
    res_trafo3w_df = net["res_trafo3w"] if suffix is None else net["res_trafo3w%s" % suffix]

    res_trafo3w_df["p_hv_mw"].values[:] = p_hv_mw
    res_trafo3w_df["q_hv_mvar"].values[:] = q_hv_mvar
    res_trafo3w_df["p_mv_mw"].values[:] = p_mv_mw
    res_trafo3w_df["q_mv_mvar"].values[:] = q_mv_mvar
    res_trafo3w_df["p_lv_mw"].values[:] = p_lv_mw
    res_trafo3w_df["q_lv_mvar"].values[:] = q_lv_mvar
    res_trafo3w_df["pl_mw"].values[:] = pl_mw
    res_trafo3w_df["ql_mvar"].values[:] = ql_mvar
    res_trafo3w_df["i_hv_ka"].values[:] = i_h
    res_trafo3w_df["i_mv_ka"].values[:] = i_m
    res_trafo3w_df["i_lv_ka"].values[:] = i_l
    res_trafo3w_df["vm_hv_pu"].values[:] = ppc["bus"][hv_buses, VM]
    res_trafo3w_df["va_hv_degree"].values[:] = ppc["bus"][hv_buses, VA]
    res_trafo3w_df["vm_mv_pu"].values[:] = ppc["bus"][mv_buses, VM]
    res_trafo3w_df["va_mv_degree"].values[:] = ppc["bus"][mv_buses, VA]
    res_trafo3w_df["vm_lv_pu"].values[:] = ppc["bus"][lv_buses, VM]
    res_trafo3w_df["va_lv_degree"].values[:] = ppc["bus"][lv_buses, VA]
    res_trafo3w_df["va_internal_degree"].values[:] = ppc["bus"][aux_buses, VA]
    res_trafo3w_df["vm_internal_pu"].values[:] = ppc["bus"][aux_buses, VM]
    res_trafo3w_df["loading_percent"].values[:] = loading_percent


def _get_impedance_results(net, ppc, i_ft, suffix=None):
    ac = net["_options"]["ac"]

    if not "impedance" in net._pd2ppc_lookups["branch"]:
        return
    f, t = net._pd2ppc_lookups["branch"]["impedance"]
    pf_mw = ppc["branch"][f:t, (PF)].real
    pt_mw = ppc["branch"][f:t, (PT)].real
    p_from_mw = pf_mw
    p_to_mw = pt_mw

    if ac:
        q_from_mvar = ppc["branch"][f:t, (QF)].real
        q_to_mvar = ppc["branch"][f:t, (QT)].real
        ql_mvar = q_from_mvar + q_to_mvar
        pl_mw = pf_mw + pt_mw
    else:
        zeros_ = np.zeros_like(p_from_mw)
        # this looks like a pyramid
        q_from_mvar = zeros_
        q_to_mvar = zeros_
        ql_mvar = zeros_
        pl_mw = zeros_
        # zeros_

    i_from_ka = i_ft[f:t][:, 0]
    i_to_ka = i_ft[f:t][:, 1]

    # write to impedance
    res_impedance_df = net["res_impedance"] if suffix is None else net["res_impedance%s" % suffix]

    res_impedance_df["p_from_mw"].values[:] = p_from_mw
    res_impedance_df["q_from_mvar"].values[:] = q_from_mvar
    res_impedance_df["p_to_mw"].values[:] = p_to_mw
    res_impedance_df["q_to_mvar"].values[:] = q_to_mvar
    res_impedance_df["pl_mw"].values[:] = pl_mw
    res_impedance_df["ql_mvar"].values[:] = ql_mvar
    res_impedance_df["i_from_ka"].values[:] = i_from_ka
    res_impedance_df["i_to_ka"].values[:] = i_to_ka


def _get_tcsc_results(net, ppc, suffix=None):
    tcsc = net.get("tcsc")  # todo: it used to be "if len(net.tcsc) == 0:"  <- change back when pandamodels bug is fixed
    if tcsc is None or len(tcsc) == 0:
        return

    ac = net["_options"]["ac"]

    baseMVA = ppc["baseMVA"]
    bus_lookup = net["_pd2ppc_lookups"]["bus"]
    f = 0
    t = len(net.tcsc)

    f_bus = bus_lookup[net.tcsc["from_bus"].values]
    t_bus = bus_lookup[net.tcsc["to_bus"].values]
    baseZ = ppc["bus"][f_bus, BASE_KV] ** 2 / baseMVA

    if ac:
        p_from_mw = ppc["tcsc"][f:t, TCSC_PF].real
        p_to_mw = ppc["tcsc"][f:t, TCSC_PT].real
        q_from_mvar = ppc["tcsc"][f:t, TCSC_QF].real
        q_to_mvar = ppc["tcsc"][f:t, TCSC_QT].real
        i_from_ka = ppc["tcsc"][f:t, TCSC_IF].real
        i_to_ka = ppc["tcsc"][f:t, TCSC_IT].real
        # losses will be 0 because tcsc represents an ideal device but we can add resistance easily in the future
        pl_mw = p_from_mw + p_to_mw
        ql_mvar = q_from_mvar + q_to_mvar
    else:
        raise NotImplementedError("TCSC not implemented for algorithm != 'ac' - writing zeros to tcsc results. "
                                  "Results for other elements are as if TCSC elements were open switches.")
        zeros_ = np.zeros(t)
        # this looks like a pyramid
        p_from_mw = zeros_
        p_to_mw = zeros_
        q_from_mvar = zeros_
        q_to_mvar = zeros_
        ql_mvar = zeros_
        pl_mw = zeros_
        i_from_ka = zeros_
        i_to_ka = zeros_
        # zeros_

    # write to impedance
    # todo for suffix not None
    res_tcsc_df = net["res_tcsc"] if suffix is None else net["res_tcsc%s" % suffix]

    res_tcsc_df["thyristor_firing_angle_degree"].values[:] = np.rad2deg(ppc["tcsc"][f:t, TCSC_THYRISTOR_FIRING_ANGLE].real)
    res_tcsc_df["x_ohm"].values[:] = ppc["tcsc"][f:t, TCSC_X_PU].real * baseZ
    res_tcsc_df["p_from_mw"].values[:] = p_from_mw
    res_tcsc_df["q_from_mvar"].values[:] = q_from_mvar
    res_tcsc_df["p_to_mw"].values[:] = p_to_mw
    res_tcsc_df["q_to_mvar"].values[:] = q_to_mvar
    res_tcsc_df["i_ka"].values[:] = np.fmax(i_from_ka, i_to_ka)
    # res_tcsc_df["i_from_ka"].values[:] = i_from_ka
    # res_tcsc_df["i_to_ka"].values[:] = i_to_ka
    res_tcsc_df["pl_mw"].values[:] = pl_mw
    res_tcsc_df["ql_mvar"].values[:] = ql_mvar
    res_tcsc_df["vm_from_pu"].values[:] = ppc["bus"][f_bus, VM]
    res_tcsc_df["va_from_degree"].values[:] = ppc["bus"][f_bus, VA]
    res_tcsc_df["vm_to_pu"].values[:] = ppc["bus"][t_bus, VM]
    res_tcsc_df["va_to_degree"].values[:] = ppc["bus"][t_bus, VA]


def _get_xward_branch_results(net, ppc, bus_lookup_aranged, pq_buses, suffix=None):
    ac = net["_options"]["ac"]

    if not "xward" in net._pd2ppc_lookups["branch"]:
        return
    f, t = net._pd2ppc_lookups["branch"]["xward"]
    p_branch_xward = ppc["branch"][f:t, PF].real
    net["res_xward"]["p_mw"].values[:] = net["res_xward"]["p_mw"].values + p_branch_xward
    if ac:
        q_branch_xward = ppc["branch"][f:t, QF].real
        net["res_xward"]["q_mvar"].values[:] = net["res_xward"]["q_mvar"].values + q_branch_xward
    else:
        q_branch_xward = np.zeros(len(p_branch_xward))
    b_pp, p, q = _sum_by_group(net["xward"]["bus"].values, p_branch_xward, q_branch_xward)
    b_ppc = bus_lookup_aranged[b_pp]

    pq_buses[b_ppc, 0] += p
    pq_buses[b_ppc, 1] += q
    aux_buses = net["_pd2ppc_lookups"]["bus"][net["_pd2ppc_lookups"]["aux"]["xward"]]

    res_xward_df = net["res_xward"] if suffix is None else net["res_xward%s" % suffix]

    res_xward_df["va_internal_degree"].values[:] = ppc["bus"][aux_buses, VA]
    res_xward_df["vm_internal_pu"].values[:] = ppc["bus"][aux_buses, VM]
    res_xward_df.index = net["xward"].index


def _get_switch_results(net, ppc, i_ft, suffix=None):
    if len(net.switch) == 0:
        return
    res_switch_df = "res_switch" if suffix is None else "res_switch%s" % suffix

    if "switch" in net._pd2ppc_lookups["branch"]:
        f, t = net._pd2ppc_lookups["branch"]["switch"]
        with np.errstate(invalid='ignore'):
            i_ka = np.max(i_ft[f:t], axis=1)
        net[res_switch_df].loc[net._impedance_bb_switches, "i_ka"] = i_ka

        p_from_mw = ppc["branch"][f:t, PF].real
        q_from_mvar = ppc["branch"][f:t, QF].real

        p_to_mw = ppc["branch"][f:t, PT].real
        q_to_mvar = ppc["branch"][f:t, QT].real

        net[res_switch_df].loc[net._impedance_bb_switches,"p_from_mw"] = p_from_mw
        net[res_switch_df].loc[net._impedance_bb_switches,"q_from_mvar"] = q_from_mvar
        net[res_switch_df].loc[net._impedance_bb_switches,"p_to_mw"] = p_to_mw
        net[res_switch_df].loc[net._impedance_bb_switches,"q_to_mvar"] = q_to_mvar

    _copy_switch_results_from_branches(net, suffix)
    if "in_ka" in net.switch.columns:
        net[res_switch_df]["loading_percent"] = net[res_switch_df]["i_ka"].values / net.switch["in_ka"].values * 100


def _copy_switch_results_from_branches(net, suffix=None, current_parameter="i_ka"):
    res_switch_df = "res_switch" if suffix is None else "res_switch%s" % suffix

    switch_lines = net.switch.element[net.switch.et=="l"]
    if len(switch_lines) > 0:
        res_line_df = "res_line" if suffix is None else "res_line%s" % suffix
        net[res_switch_df].loc[switch_lines.index, current_parameter] = net[res_line_df].loc[switch_lines.values, current_parameter].values

    switch_trafo = net.switch[net.switch.et.values=="t"]
    if len(switch_trafo) > 0:
        res_trafo_df = "res_trafo" if suffix is None else "res_trafo%s" % suffix
        for side in ["hv", "lv"]:
            buses = net.trafo["{}_bus".format(side)].loc[switch_trafo.element.values].values
            side_switch_trafo = switch_trafo[switch_trafo.bus.values==buses]
            switches = side_switch_trafo.index
            trafos = side_switch_trafo.element.values
            current, unit = current_parameter.split("_")
            side_current = "{}_{}_{}".format(current, side, unit)
            net[res_switch_df].loc[switches, current_parameter] = net[res_trafo_df].loc[trafos, side_current].values
    open_switches = ~net.switch.closed.values
    if any(open_switches):
        net[res_switch_df].loc[open_switches, current_parameter] = 0



from copy import deepcopy
from functools import reduce
import operator
import numpy as np
import pandas as pd

from pandapower.auxiliary import ensure_iterability
from pandapower.toolbox.element_selection import pp_elements, element_bus_tuples
from pandapower.topology.create_graph import create_nxgraph
from pandapower.topology.graph_searches import connected_component, connected_components


try:
    from pandaplan.core import pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)


def getFromDict(dict_, keys):
    """ Get value from nested dict """
    return reduce(operator.getitem, keys, dict_)


def setInDict(dict_, keys, value):
    """ Set value to nested dict """
    getFromDict(dict_, keys[:-1])[keys[-1]] = value


def appendSetInDict(dict_, keys, set_):
    """ Use case specific: append existing value of type set in nested dict """
    getFromDict(dict_, keys[:-1])[keys[-1]] |= set_


def setSetInDict(dict_, keys, set_):
    """ Use case specific: set new or append existing value of type set in nested dict """
    if isinstance(getFromDict(dict_, keys[:-1]), dict):
        if keys[-1] in getFromDict(dict_, keys[:-1]).keys():
            if isinstance(getFromDict(dict_, keys), set):
                appendSetInDict(dict_, keys, set_)
            else:
                raise ValueError("The set in the nested dict cannot be appended since it actually "
                                 "is not a set but a " + str(type(getFromDict(dict_, keys))))
        else:
            setInDict(dict_, keys, set_)
    else:
        raise ValueError("This function expects a dict for 'getFromDict(dict_, " + str(keys[:-1]) +
                         ")', not a" + str(type(getFromDict(dict_, keys[:-1]))))


def append_set_to_dict(dict_, set_, keys):
    """ Appends a nested dict by the values of a set, independant if the keys already exist or not.
    """
    keys = ensure_iterability(keys)

    # ensure that the dict way to the last key exist
    for pos, _ in enumerate(keys[:-1]):
        if isinstance(getFromDict(dict_, keys[:pos]), dict):
            if keys[pos] not in getFromDict(dict_, keys[:pos]).keys():
                setInDict(dict_, keys[:pos + 1], dict())
        else:
            raise ValueError("This function expects a dict for 'getFromDict(dict_, " +
                             str(keys[:pos]) + ")', not a" + str(type(getFromDict(
                dict_, keys[:pos]))))

    # set the value
    setSetInDict(dict_, keys, set_)


def set_bus_zone_by_boundary_branches(net, all_boundary_branches):
    """
    Set integer values (0, 1, 2, ...) to net.bus.zone with regard to the given boundary branches in
    'all_boundary_branches'.

    INPUT:
        **net** - pandapowerNet

        **all_boundary_branches** (dict) - defines which element indices are boundary branches.
            The dict keys must be pandapower elements, e.g. "line" or "trafo"
    """
    include = dict.fromkeys(["line", "trafo", "trafo3w", "impedance"])
    for elm in include.keys():
        if elm in all_boundary_branches.keys():
            include[elm] = net[elm].index.difference(all_boundary_branches[elm])
        else:
            include[elm] = True

    mg = create_nxgraph(net, include_lines=include["line"], include_impedances=include[
        "impedance"], include_trafos=include["trafo"], include_trafo3ws=include["trafo3w"])
    cc = connected_components(mg)
    ccl = [set_ for set_ in cc]
    areas = []

    while len(ccl):
        # check intersections of the first area with all other unchecked areas (remains in ccl) and
        # then add first area unionized with all intersectioned other areas to "areas"
        areas += [ccl.pop(0)]
        n_last_area = -1
        while n_last_area != len(areas[-1]):
            # check as long as len(areas[-1]) not changes anymore - needed because there can be
            # intersections of remaining areas with the buses added to areas[-1]
            # within the last while loop iteration via union
            n_last_area = len(areas[-1])
            for i, set_ in enumerate(ccl):
                if set_.intersection(areas[-1]):
                    areas[-1] |= ccl.pop(i)

    for i, area in enumerate(areas):
        net.bus.loc[list(area), "zone"] = i


def get_boundaries_by_bus_zone_with_boundary_branches(net):
    """
    Only in_service branches and closed switches are considered.

    INPUT:
        **net** - pandapower net - In net.bus.zone different zones must be given.

    OUTPUT:
        **boundary_buses** - dict of boundary buses - for each zone the internal and external
        boundary buses are given as well as the external boundary buses for each other zone.
        Furthermore the value of the boundary_buses key "all" concludes all boundary buses of all
        zones.
        Example:
            {"all": {0, 1, 3},
             1: {"all": {0, 1, 3},
                 "external": {3},
                 "internal": {0, 1},
                 2: {3}
                 },
             2: {"all": {1, 2, 3, 4},
                 "external": {0, 1, 4},
                 "internal": {2, 3},
                 1: {0, 1},
                 3: {4}
                 },
             3: {"all": {2, 4},
                 "external": {2},
                 "internal": {4},
                 2: {2}
                 }
             }

        **boundary_branches** - dict of branch elements - for each zone a set of the corresponding
        boundary boundary branches as well as "all" boundary branches
        Example:
            {"all": {"line": {0, 1},
                     "trafo": {0}
                     },
             1: {"line": {0},
                 "trafo": {0}
                 },
             2: {"line": {0, 1}},
             3: {"line": {1}}
             }
    """

    def append_boundary_buses_externals_per_zone(boundary_buses, boundaries, zone, other_zone_cols):
        """ iterate throw all boundaries which matches this_zone and add the other_zone_bus to
        boundary_buses """
        for idx, ozc in other_zone_cols.items():
            other_zone = boundaries[zone_cols].values[idx, ozc]
            if isinstance(other_zone, np.generic):
                other_zone = other_zone.item()
            if zone == other_zone:
                continue  # this happens if 2 trafo3w connections are in zone
            other_zone_bus = boundaries[buses].values[idx, ozc]
            append_set_to_dict(boundary_buses, {other_zone_bus}, [zone, other_zone])

    if "all" in set(net.bus.zone.values):
        raise ValueError("'all' is not a proper zone name.")  # all is used later for other purpose
    branch_elms = pp_elements(bus=False, bus_elements=False, branch_elements=True,
                              other_elements=False, res_elements=False)
    branch_tuples = element_bus_tuples(bus_elements=False, branch_elements=True,
                                       res_elements=False) + [("switch", "element")]
    branch_dict = {branch_elm: [] for branch_elm in branch_elms}
    for elm, bus in branch_tuples:
        branch_dict[elm] += [bus]

    zones = net.bus.zone.unique()
    boundary_branches = {zone if net.bus.zone.dtype == object else zone.item():
                             dict() for zone in zones}
    boundary_branches["all"] = dict()
    boundary_buses = {zone if net.bus.zone.dtype == object else zone.item():
                          {"all": set(), "internal": set(), "external": set()} for zone in zones}
    boundary_buses["all"] = set()

    for elm, buses in branch_dict.items():
        idx = net[elm].index[net[elm].in_service] if elm != "switch" else net.switch.index[
            (net.switch.et == "b") & net.switch.closed]
        boundaries = deepcopy(net[elm][buses].loc[idx])  # copy relevant info from net[elm]

        zone_cols = list()
        for i, bus_col in enumerate(buses):
            # add to which zones the connected buses belong to
            boundaries["zone%i" % i] = net.bus.zone.loc[boundaries[bus_col].values].values
            zone_cols.append("zone%i" % i)
            # compare the zones and conclude if the branches can be boundaries
            if i > 0:
                boundaries["is_boundary"] |= boundaries["zone%i" % i] != boundaries["zone0"]
            else:
                boundaries["is_boundary"] = False
        # reduce the DataFrame 'boundaries' to those branches which actually are boundaries
        boundaries = boundaries.loc[boundaries["is_boundary"],
        boundaries.columns.difference(["is_boundary"])]

        # determine boundary_branches and boundary_buses
        if len(boundaries):
            boundary_branches["all"][elm] = set(boundaries.index)
            boundary_buses["all"] |= set(boundaries[buses].values.flatten())

            for zone in set(boundaries[zone_cols].values.flatten()):

                # determine which columns belong to this zone and which not
                this_zone_col = np.zeros(boundaries.shape[0]) * np.nan
                for i, _ in enumerate(buses):
                    this_zone_col[boundaries[zone_cols[i]] == zone] = i
                this_zone_col = pd.Series(this_zone_col).dropna().astype(np.int64)
                other_zone_col1 = pd.Series(np.ones(this_zone_col.shape, dtype=np.int64),
                                            index=this_zone_col.index) - this_zone_col
                if len(buses) == 3:
                    other_zone_col1.loc[other_zone_col1 < 0] = 0
                    other_zone_col2 = pd.Series(3 * np.ones(this_zone_col.shape, dtype=np.int64),
                                                index=this_zone_col.index) - \
                                      this_zone_col - other_zone_col1

                # fill zone dependant values to boundary_branches and boundary_buses
                boundary_branches[zone][elm] = set(boundaries.index[this_zone_col.index])

                nint = set(boundaries[buses].values[this_zone_col.index, this_zone_col.values])
                ext = set(boundaries[buses].values[other_zone_col1.index, other_zone_col1.values])
                boundary_buses[zone]["internal"] |= nint
                boundary_buses[zone]["external"] |= ext
                boundary_buses[zone]["all"] |= ext | nint
                if len(buses) == 3:
                    ext = set(boundaries[buses].values[
                                  other_zone_col2.index, other_zone_col2.values])
                    boundary_buses[zone]["external"] |= ext
                    boundary_buses[zone]["all"] |= ext

                append_boundary_buses_externals_per_zone(
                    boundary_buses, boundaries, zone, other_zone_col1)
                if len(buses) == 3:
                    append_boundary_buses_externals_per_zone(
                        boundary_buses, boundaries, zone, other_zone_col2)

    # check for missing zone connections
    zones_without_connection = list()
    for zone, bra in boundary_branches.items():
        if zone != "all" and not bra:
            zones_without_connection.append(zone)
    if len(zones_without_connection):
        logger.warning("These zones have no connections to other zones: " + str(
            zones_without_connection))

    return boundary_buses, boundary_branches


def get_connected_switch_buses_groups(net, buses):
    all_buses = set()
    bus_dict = []
    mg_sw = create_nxgraph(net, include_trafos=False,
                           include_trafo3ws=False,
                           respect_switches=True,
                           include_lines=False,
                           include_impedances=False)
    for bbus in buses:
        if bbus in all_buses:
            continue
        new_bus_set = set(connected_component(mg_sw, bbus))
        all_buses |= new_bus_set
        bus_dict.append(list(new_bus_set))
    return all_buses, bus_dict


if __name__ == "__main__":
    pass



import logging
import time

import pandas as pd
import numpy as np

from pandapower.converter.cim import cim_tools
from pandapower.converter.cim.cim2pp import build_pp_net
from pandapower.converter.cim.other_classes import Report, LogLevel, ReportCode

logger = logging.getLogger('cim.cim2pp.converter_classes.acLineSegmentsCim16')

sc = cim_tools.get_pp_net_special_columns_dict()


class AcLineSegmentsCim16:
    def __init__(self, cimConverter: build_pp_net.CimConverter):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.cimConverter = cimConverter

    def convert_ac_line_segments_cim16(self, convert_line_to_switch, line_r_limit, line_x_limit):
        time_start = time.time()
        self.logger.info("Start converting ACLineSegments.")
        eq_ac_line_segments = self._prepare_ac_line_segments_cim16(convert_line_to_switch, line_r_limit, line_x_limit)

        # now create the lines and the switches
        # -------- lines --------
        if 'line' in eq_ac_line_segments['kindOfType'].values:
            line_df = eq_ac_line_segments.loc[eq_ac_line_segments['kindOfType'] == 'line']
            line_df = line_df.rename(columns={
                'rdfId': sc['o_id'], 'rdfId_Terminal': sc['t_from'], 'rdfId_Terminal2': sc['t_to'],
                'index_bus': 'from_bus', 'index_bus2': 'to_bus', 'length': 'length_km',
                'shortCircuitEndTemperature': 'endtemp_degree', 'EquipmentContainer': 'EquipmentContainer_id'})
            line_df[sc['o_cl']] = 'ACLineSegment'
            line_df['in_service'] = line_df.connected & line_df.connected2
            line_df['r_ohm_per_km'] = abs(line_df.r) / line_df.length_km
            line_df['x_ohm_per_km'] = abs(line_df.x) / line_df.length_km
            line_df['c_nf_per_km'] = abs(line_df.bch) / (2 * 50 * np.pi * line_df.length_km) * 1e9
            line_df['g_us_per_km'] = abs(line_df.gch) * 1e6 / line_df.length_km
            line_df['r0_ohm_per_km'] = abs(line_df.r0) / line_df.length_km
            line_df['x0_ohm_per_km'] = abs(line_df.x0) / line_df.length_km
            line_df['c0_nf_per_km'] = abs(line_df.b0ch) / (2 * 50 * np.pi * line_df.length_km) * 1e9
            line_df['g0_us_per_km'] = abs(line_df.g0ch) * 1e6 / line_df.length_km
            line_df['parallel'] = 1
            line_df['df'] = 1.
            line_df['type'] = None
            line_df['std_type'] = None
            self.cimConverter.copy_to_pp('line', line_df)
        else:
            line_df = pd.DataFrame(None)
        # -------- switches --------
        if 'switch' in eq_ac_line_segments['kindOfType'].values:
            switch_df = eq_ac_line_segments.loc[eq_ac_line_segments['kindOfType'] == 'switch']

            switch_df = switch_df.rename(columns={
                'rdfId': sc['o_id'], 'index_bus': 'bus', 'index_bus2': 'element', 'rdfId_Terminal': sc['t_bus'],
                'rdfId_Terminal2': sc['t_ele']})
            switch_df['et'] = 'b'
            switch_df['type'] = None
            switch_df['z_ohm'] = 0
            if switch_df.index.size > 0:
                switch_df['closed'] = switch_df.connected & switch_df.connected2
            self.cimConverter.copy_to_pp('switch', switch_df)
        else:
            switch_df = pd.DataFrame(None)

        self.logger.info("Created %s lines and %s switches in %ss" %
                         (line_df.index.size, switch_df.index.size, time.time() - time_start))
        self.cimConverter.report_container.add_log(Report(
            level=LogLevel.INFO, code=ReportCode.INFO_CONVERTING,
            message="Created %s lines and %s switches from ACLineSegments in %ss" %
                    (line_df.index.size, switch_df.index.size, time.time() - time_start)))

    def _prepare_ac_line_segments_cim16(self, convert_line_to_switch, line_r_limit, line_x_limit) -> pd.DataFrame:

        if 'sc' in self.cimConverter.cim.keys():
            ac_line_segments = self.cimConverter.merge_eq_sc_profile('ACLineSegment')
        else:
            ac_line_segments = self.cimConverter.cim['eq']['ACLineSegment']

        line_length_before_merge = ac_line_segments.index.size
        # until now self.cim['eq']['ACLineSegment'] looks like:
        #   rdfId   name    r       ...
        #   _x01    line1   0.056   ...
        #   _x02    line2   0.471   ...
        # now join with the terminals
        ac_line_segments = pd.merge(ac_line_segments, self.cimConverter.bus_merge,
                                       how='left', on='rdfId')
        ac_line_segments[sc['o_cl']] = 'ACLineSegment'
        # now ac_line_segments looks like:
        #   rdfId   name    r       rdfId_Terminal  connected   ...
        #   _x01    line1   0.056   termi025        True        ...
        #   _x01    line1   0.056   termi223        True        ...
        #   _x02    line2   0.471   termi154        True        ...
        #   _x02    line2   0.471   termi199        True        ...
        # if each switch got two terminals, reduce back to one line to use fast slicing
        if ac_line_segments.index.size != line_length_before_merge * 2:
            self.logger.error("Error processing the ACLineSegments, there is a problem with Terminals in the source "
                              "data!")
            self.cimConverter.report_container.add_log(Report(
                level=LogLevel.ERROR, code=ReportCode.ERROR_CONVERTING,
                message="Error processing the ACLineSegments, there is a problem with Terminals in the source data!"))
            dups = ac_line_segments.pivot_table(index=['rdfId'], aggfunc='size')
            dups = dups.loc[dups != 2]
            for rdfId, count in dups.items():
                self.logger.warning("The ACLineSegment with RDF ID %s has %s Terminals!" % (rdfId, count))
                self.logger.warning("The ACLineSegment data: \n%s" %
                                    ac_line_segments[ac_line_segments['rdfId'] == rdfId])
                self.cimConverter.report_container.add_log(Report(
                    level=LogLevel.WARNING, code=ReportCode.WARNING_CONVERTING,
                    message="The ACLineSegment with RDF ID %s has %s Terminals!" % (rdfId, count)))
            ac_line_segments = ac_line_segments[0:0]
        ac_line_segments = ac_line_segments.reset_index()
        # now merge with OperationalLimitSets and CurrentLimits
        eq_operational_limit_sets = self.cimConverter.cim['eq']['OperationalLimitSet'][['rdfId', 'Terminal']]
        eq_operational_limit_sets = eq_operational_limit_sets.rename(columns={'rdfId': 'rdfId_OperationalLimitSet',
                                                  'Terminal': 'rdfId_Terminal'})
        ac_line_segments = pd.merge(ac_line_segments, eq_operational_limit_sets, how='left',
                                       on='rdfId_Terminal')
        if 'CurrentLimit' in self.cimConverter.cim['ssh'].keys():
            current_limits = self.cimConverter.merge_eq_ssh_profile('CurrentLimit')[['rdfId', 'OperationalLimitSet',
                                                                                     'value']]
        else:
            current_limits = self.cimConverter.cim['eq']['CurrentLimit'][['rdfId', 'OperationalLimitSet', 'value']]
        current_limits = current_limits.rename(columns={'rdfId': 'rdfId_CurrentLimit',
                                          'OperationalLimitSet': 'rdfId_OperationalLimitSet'})
        ac_line_segments = pd.merge(ac_line_segments, current_limits, how='left',
                                    on='rdfId_OperationalLimitSet')
        ac_line_segments.value = ac_line_segments.value.astype(float)
        # sort by rdfId, sequenceNumber and value. value is max_i_ka, choose the lowest one if more than one is
        # given (A line may have more than one max_i_ka in CIM, different modes e.g. normal)
        ac_line_segments = ac_line_segments.sort_values(by=['rdfId', 'sequenceNumber', 'value'])
        ac_line_segments = ac_line_segments.drop_duplicates(['rdfId', 'rdfId_Terminal'], keep='first')

        # copy the columns which are needed to reduce the ac_line_segments to one row per line
        ac_line_segments['rdfId_Terminal2'] = ac_line_segments['rdfId_Terminal'].copy()
        ac_line_segments['connected2'] = ac_line_segments['connected'].copy()
        ac_line_segments['index_bus2'] = ac_line_segments['index_bus'].copy()
        ac_line_segments['value2'] = ac_line_segments['value'].copy()
        ac_line_segments = ac_line_segments.reset_index()
        # here is where the magic happens: just remove the first value from the copied columns, reset the index
        # and replace the old column with the cut one. At least just remove the duplicates on column rdfId
        ac_line_segments.rdfId_Terminal2 = ac_line_segments.rdfId_Terminal2.iloc[
                                              1:].reset_index().rdfId_Terminal2
        ac_line_segments.connected2 = ac_line_segments.connected2.iloc[1:].reset_index().connected2
        ac_line_segments.index_bus2 = ac_line_segments.index_bus2.iloc[1:].reset_index().index_bus2
        ac_line_segments.value2 = ac_line_segments.value2.iloc[1:].reset_index().value2
        ac_line_segments = ac_line_segments.drop_duplicates(['rdfId'], keep='first')
        # get the max_i_ka
        ac_line_segments['max_i_ka'] = ac_line_segments['value'].fillna(ac_line_segments['value2']) * 1e-3

        # filter if line or switches will be added
        ac_line_segments['kindOfType'] = 'line'
        if convert_line_to_switch:
            ac_line_segments.loc[(abs(ac_line_segments['r']) <= line_r_limit) |
                                    (abs(ac_line_segments['x']) <= line_x_limit), 'kindOfType'] = 'switch'
        return ac_line_segments



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

import numpy as np
from numpy import interp
from scipy.interpolate import interp1d, PchipInterpolator

from pandapower.io_utils import JSONSerializableClass

try:
    import pandaplan.core.pplog as pplog
except ImportError:
    import logging as pplog

logger = pplog.getLogger(__name__)


class Characteristic(JSONSerializableClass):
    """
    This class represents a characteristics curve. The curve is described as a piecewise linear function.

    INPUT:
        **pts** - Expects two (or more) points of the function (i.e. kneepoints)

    OPTIONAL:
        **eps** - An epsilon to compare the difference to

    The class has an implementation of the ``__call__`` method, which allows using it interchangeably with other interpolator objects,
    e.g. ``scipy.interpolate.interp1d``, ``scipy.interpolate.CubicSpline``, ``scipy.interpolate.PPoly``, etc.

    Example usage:
        Create a simple function from two points and ask for the target y-value for a
        given x-value.
        Assume a characteristics curve in which for voltages < 0.95pu a power of 10kW
        is desired, linear rising to a max. of 20 kW at 1.05 pu

    ::

        You can give points by lists of x/y-values
        >>> c = Characteristic(net, x_values=[0.95, 1.05],y_values=[10, 20])
        >>> c(x=1.0)
        15.0

        or pass a list of points (x,y)
        >>> c = Characteristic.from_points(net,points=[(0.95, 10), (1.05, 20)])
        >>> c(x=1.0)
        15.0

        or in a simple case from a gradient, its zero crossing and the maximal values for y
        >>> c = Characteristic.from_gradient(net,zero_crossing=-85,gradient=100,y_min=10,y_max=20)
        >>> c(x=1.0)
        15.0

        Values are constant beyond the first and last defined points
        >>> c(x=42)
        20.0
        >>> c(x=-42)
        10.0

        Create a curve with many points and ask for the difference between the y-value being measured
        and the expected y-value for a given x-value
        >>> c = Characteristic.from_points(net,points=[(1,2),(2,4),(3,2),(42,24)])
        >>> c.diff(x=2.5, measured=3)
        0.0

        You can also ask if a y-values satisfies the curve at a certain x-value. Note how the use of
        an epsilon behaves (for x=2.5 we expect 3.0):
        >>> c.satisfies(x=2.5, measured=3.099999999, epsilon=0.1)
        True
        >>> c.satisfies(x=2.5, measured=3.1, epsilon=0.1)
        False
    """

    def __init__(self, net, x_values, y_values, table="characteristic", **kwargs):
        super().__init__()
        self.x_vals = x_values
        self.y_vals = y_values
        self.index = super().add_to_net(net, table)

    # @property
    # def x_vals(self):
    #     return self._x_vals
    #
    # @property
    # def y_vals(self):
    #     return self._y_vals

    @classmethod
    def from_points(cls, net, points, **kwargs):
        unzipped = list(zip(*points))
        return cls(net, unzipped[0], unzipped[1], **kwargs)

    @classmethod
    def from_gradient(cls, net, zero_crossing, gradient, y_min, y_max, **kwargs):
        x_left = (y_min - zero_crossing) / float(gradient)
        x_right = (y_max - zero_crossing) / float(gradient)
        return cls(net, [x_left, x_right], [y_min, y_max], **kwargs)

    def diff(self, x, measured):
        """

        INPUT:
            **x** - The x-value at which the current y-value is measured
            **actual** - The actual y-value being measured.
            **return** - The difference between actual and expected value.
        """
        return measured - self(x)

    def satisfies(self, x, measured, epsilon):
        """

        INPUT:
            **x** - The x-value at which the current y-value is measured

            **measured** - The actual y-value being measured.

        OUTPUT:
            Whether or not the point satisfies the characteristics curve with respect to the
            epsilon being set
        """
        if abs(self.diff(x, measured)) < epsilon:
            return True
        else:
            return False

    def __call__(self, x):
        """

        INPUT:
            **x** - An x-value

        OUTPUT:
            The corresponding target value of this characteristics
        """
        return interp(x, self.x_vals, self.y_vals)

    def __repr__(self):
        return self.__class__.__name__


class SplineCharacteristic(Characteristic):
    """
    SplineCharacteristic interpolates the y-value(s) for the given x-value(s) according to a non-linear function.
    Internally the interpolator object interp1d from scipy.interpolate is used.
    By default, the function is quadratic, but the user can specify other methods (refer to the documentation of
    interp1d). The fill_value can be specified as "extrapolate" so that even x-values outside of the specified
    range can be used and yield y-values outside the specified y range. Alternatively, the behavior of
    Characteristic can be followed by providing a tuple for the fill value for x outside the specified range,
    refer to the documentation of interp1d for more details. We set the parameter bounds_error to False.

    INPUT:
        **net**

        **x_values**

        **y_values**

        **fill_value**
    """
    json_excludes = ["self", "__class__", "_interpolator"]

    def __init__(self, net, x_values, y_values, interpolator_kind="interp1d", table="characteristic", **kwargs):
        super().__init__(net, x_values=x_values, y_values=y_values, table=table)
        self.kwargs = kwargs
        self.interpolator_kind = interpolator_kind

    @property
    def interpolator(self):
        """
        We need to store the interpolator in a property because we need to serialize
        the characteristic. Instead of storing the serialized interpolator, we store the
        x_values and y_values (the attribute _interpolator is excluded from serialization by
        adding it to json_excludes). For it to work, we need to recreate the interpolator on
        demand. As soon as the characteristic is called, if the interpolator is there,
        we can use it. If not, we recreate it.
        """
        return self._interpolator

    @interpolator.getter
    def interpolator(self):
        if not hasattr(self, '_interpolator'):
            if self.interpolator_kind == "interp1d":
                self._interpolator = default_interp1d(self.x_vals, self.y_vals, **self.kwargs)
            elif self.interpolator_kind == "Pchip":
                self._interpolator = PchipInterpolator(self.x_vals, self.y_vals, **self.kwargs)
            else:
                raise NotImplementedError(f"Interpolator {self.interpolator_kind} not implemented!")
        return self._interpolator

    def __call__(self, x):
        """
        This method allows calling the SciPy interpolator object directly.
        Codacy is complaining about this, but it is not a problem.

        INPUT:
            **x** (float) -  The x-value at which the current y-value is interpolated for.

        OUTPUT:
            The interpolated y-value.
        """
        return self.interpolator(x)


class LogSplineCharacteristic(SplineCharacteristic):

    def __init__(self, net, x_values, y_values, **kwargs):
        super().__init__(net, x_values, y_values, **kwargs)

    @property
    def x_vals(self):
        return self._x_vals

    @property
    def y_vals(self):
        return self._y_vals

    @x_vals.setter
    def x_vals(self, x_values):
        if np.any(x_values == 0):
            logger.warning("zero-values not supported in x_values")
        self._x_vals = np.log10(x_values)

    @y_vals.setter
    def y_vals(self, y_values):
        if np.any(y_values == 0):
            logger.warning("zero-values not supported in y_values")
        self._y_vals = np.log10(y_values)

    def __call__(self, x):
        return np.power(10, self.interpolator(np.log10(x)))


def default_interp1d(x, y, kind="quadratic", bounds_error=False, fill_value="extrapolate", **kwargs):
    return interp1d(x, y, kind=kind, bounds_error=bounds_error, fill_value=fill_value, **kwargs)



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

from pandapower.control.controller.const_control import ConstControl
try:
    from pandaplan.core import pplog
except:
    import logging as pplog
import numpy as np

logger = pplog.getLogger(__name__)


class PQController(ConstControl):
    """
    Parent class of all Controllers designed to control P and/or Q of an PQ-element.
    Can be used for loads, shunts and sgen.
    NOT applicable for gens.
    Contains general initialization and
    functions for getting and setting values from/to the data-structure.
    Furthermore it contains the convergence check in respect to the
    criteria it was created with.

    INPUT:
        **net** (attrdict) - pandapower network

        **element_index** (int[]) - IDs of the controlled elements

    OPTIONAL:

        **element** - element table ('sgen', 'load' etc.)

        **max_p_error** (float, 0.0001) - Maximum absolute error of active power in MW

        **max_q_error** (float, 0.0001) - Maximum absolute error of reactive power in Mvar

        **pq_simultaneity_factor** (float, 1.0) - Simultaneity factor applied to P and Q

        **converter_sizing_pu** (float, 1.0) - Sizing of the converter factor applied to limit the
        active power (using the internal function limit_to_inverter_sizing())

        **data_source** ( , None) - A DataSource that contains profiles

        **profile_scale** (float, 1.0) - A scaling factor applied to the values of profiles

        **in_service** (bool, True) - Indicates if the controller is currently in_service

        **ts_absolute** (bool, True) - Whether the time step values are absolute power values or
        scaling factors

    """

    def __init__(self, net, element_index, element="sgen", max_p_error=0.0001, max_q_error=0.0001,
                 pq_simultaneity_factor=1., converter_sizing_pu=1., data_source=None, profile_scale=1., in_service=True,
                 ts_absolute=True, order=0, level=0, **kwargs):
        super().__init__(net, element=element, variable="p_mw", element_index=element_index,
                         in_service=in_service, order=order, level=level, **kwargs)

        # read attributes from net
        self.element_index = element_index
        self.element = element
        self.bus = net[self.element]["bus"][element_index]
        self.p_mw = net[self.element]["p_mw"][element_index]
        self.q_mvar = net[self.element]["q_mvar"][element_index]
        self.sn_mva = net[self.element]["sn_mva"][element_index]
        self.element_names = net[self.element]["name"][element_index]
        self.gen_type = net[self.element]["type"][element_index]
        self.element_in_service = net[self.element]["in_service"][element_index]

        self.sign = 1
        if element == "sgen":
            self.sign *=-1
        # Simultaneity factor
        self.pq_simultaneity_factor = pq_simultaneity_factor

        # Sizing of the AC/DC converter
        self.converter_sizing_pu = converter_sizing_pu

        # attributes for profile
        self.data_source = data_source
        self.profile_scale = profile_scale
        self.p_profile = None
        self.q_profile = None
        self.ts_absolute = ts_absolute

        # Init variables for convergence check
        self.max_p_error = max_p_error
        self.max_q_error = max_q_error

        # Init curtailment
        self.p_curtailment = 0

        self.set_recycle(net)

    def set_p_profile(self, p_profile, profile_from_name):
        if profile_from_name:
            if p_profile:
                logger.warning("Given parameter 'p_profile' will be discarded "
                               "since 'profile_from_name' has been set to True.")
            self.p_profile = [f"P_{name}" for name in self.element_names]
        else:
            self.p_profile = p_profile

    def set_q_profile(self, q_profile, profile_from_name):
        if profile_from_name:
            if q_profile:
                logger.warning("Given parameter 'q_profile' will be discarded "
                               "since 'profile_from_name' has been set to True.")
            self.q_profile = [f"Q_{name}" for name in self.element_names]
        else:
            self.q_profile = q_profile

    def __repr__(self):
        rep = super(PQController, self).__repr__()

        for member in ["element_index", "bus", "p_mw", "q_mvar", "sn_mva", "element_names",
                       "gen_type", "element_in_service", "pq_simultaneity_factor",
                       "converter_sizing_pu", "data_source", "profile_scale", "p_profile", "q_profile",
                       "ts_absolute", "max_p_error",
                       "max_q_error"]:
            rep += ("\n" + member + ":").ljust(20)
            d = locals()
            exec('value = self.' + member, d)
            rep += str(d['value'])

        return rep

    def read_profiles(self, time):
        """
        Reads new data if a DataSource and profiles have been set.
        The simultaneity factor pq_simultaneity_factor is being applied directly to the value
        retrieved from the DataSource. self.profile_scale in turn is being
        passed to get_time_step_value() and applied by the DataSource.
        """

        if self.data_source is not None:
            if self.p_profile is not None:
                self.p_mw = self.pq_simultaneity_factor * \
                            self.data_source.get_time_step_value(time_step=time,
                                                                 profile_name=self.p_profile,
                                                                 scale_factor=self.profile_scale)

            if self.q_profile is not None:
                self.q_mvar = self.pq_simultaneity_factor * \
                              self.data_source.get_time_step_value(time_step=time,
                                                                   profile_name=self.q_profile,
                                                                   scale_factor=self.profile_scale)

            if not self.ts_absolute:
                if self.sn_mva.isnull().any():
                    logger.error(f"There are PQ controlled elements with NaN sn_mva values.")
                self.p_mw = self.p_mw * self.sn_mva
                self.q_mvar = self.q_mvar * self.sn_mva

            # store provided power until controller converged and subtract in finalize_step
            self.p_curtailment = self.p_mw

    def write_to_net(self, net):
        # write p, q to bus within the net
        net[self.element].loc[self.element_index, "p_mw"] = self.p_mw
        net[self.element].loc[self.element_index, "q_mvar"] = self.q_mvar

    def finalize_control(self, net):
        self.calc_curtailment()

    def calc_curtailment(self):
        # p_curtailment contains the provided power up to here,
        # now the actual power is subtracted to store the curtailment
        try:
            self.p_curtailment -= self.p_mw
        except AttributeError:
            logger.error("No p_kw present at %s. Assuming no curtailment." % self)

    def limit_to_inverter_sizing(self, p_mw, q_mvar):
        """
        Returns the limited P and Q values in respect to the SGENs inverter sizing

        INPUT:
            **p_mw** - Active power of [self.element]

            **q_mvar** - Reactive power of [self.element]

        OUTPUT:

            **p_mw** - Active power limited to inverter sizing

            **q_mvar** - Reactive power limited to inverter sizing
        """
        if n_nan_sn := sum(self.sn_mva.isnull()):
            logger.warning(f"Limiting to inverter size will result in NaN for the {n_nan_sn} PQ "
                           "controlled elements with NaN sn_mva values.")
        # limit output to inverter sizing
        if (np.sqrt(p_mw ** 2 + q_mvar ** 2) > self.converter_sizing_pu * self.sn_mva).any():
           # limit Q first
            # t = q_mvar
            # q_mvar = sign(q_mvar) * math.sqrt((self.converter_sizing_pu * self.sn_mva) ** 2 - p_mw ** 2)
            sn_mva = self.converter_sizing_pu *self.sn_mva
            max_q_mvar = np.sqrt(sn_mva ** 2 - p_mw ** 2)
            q_mvar = np.min((q_mvar*self.sign, max_q_mvar*self.sign))

            logger.debug("Note: Q_mvar has been limited"
                             "respect to the inverter sizing.")
            if np.isnan(max_q_mvar).any():  # limit P if it already exceeds Sn
                q_mvar = 0
                p_mw = self.converter_sizing_pu * self.sn_mva
                logger.debug("Note: P_mw has been limited and"
                             "q_mvar has been set to 0 in respect to sn_mva ")
        return p_mw, q_mvar

    def initialize_control(self, net):
        """
        Reading the actual P and Q state from the respective [self.element] in the net
        """
        self.p_mw = net[self.element].loc[self.element_index, "p_mw"]
        self.q_mvar = net[self.element].loc[self.element_index, "q_mvar"]

    def __str__(self):
        if len(self.element_index) > 6:
            return f"PQController(len(element_index)={len(self.element_index)}"
        else:
            return f"PQController(element_index={self.element_index}"
